from abc import abstractmethod

from bclearer_orchestration_services.reporting_service.wrappers.run_and_log_function_wrapper import (
    run_and_log_function,
)

# #from bie_base.b_source.common.configurations.b_unit_domain_configurations import BUnitDomainConfigurations
# #from core_graph_mvp_base.b_source.neo4j_entification_pipeline.objects.universes.neo4j_import_spoke_bclearer_run_universes import Neo4JImportSpokeBClearerRunUniverses
# from core_graph_mvp_base.b_source.neo4j_entification_pipeline.operations.b_units.registrations.output_domain_dataframes_registerer import register_output_domain_dataframes_in_domain_universe
# from core_graph_mvp_base.b_source.neo4j_entification_pipeline.objects.enums.reportable.bclearer_neo4j_entification_bunits import \
#     BClearerNeo4jEntificationBunits
# from core_graph_mvp_base.b_source.neo4j_entification_pipeline.operations.b_units.registrations.in_parallel_bie_universe_registerer import register_in_parallel_bie_universe
# from bie_base.b_source.common.enums.bie_enums import BieEnums
# # from bie_base.b_source.uncommon.bclearer_processing.registrations.b_unit_input_output_lists import \
# #     BUnitInputOutputLists


class BUnits:
    def __init__(
        self,
        # neo4j_import_spoke_bclearer_run_universe: Neo4JImportSpokeBClearerRunUniverses,
        # b_unit_bie_enum: BieEnums,
        # parent_bie_pipeline_component_enum: BieEnums
    ):
        # self.neo4j_import_spoke_bclearer_run_universe = \
        #     neo4j_import_spoke_bclearer_run_universe

        # self.b_unit_domain_configuration = \
        #     BUnitDomainConfigurations(
        #         b_unit_bie_enum=b_unit_bie_enum,
        #         parent_bie_pipeline_component_enum=parent_bie_pipeline_component_enum)
        pass

    def run(self) -> None:
        function_discriminator_name = (
            self.b_unit_domain_configuration.b_unit_bie_enum.b_enum_item_name
        )

        self.__run_b_unit(
            function_discriminator_name=function_discriminator_name
        )

    @run_and_log_function()
    def __run_b_unit(self) -> None:
        self.b_unit_process_function()

        # register_output_domain_dataframes_in_domain_universe(
        #     neo4j_import_spoke_bclearer_run_universe=self.neo4j_import_spoke_bclearer_run_universe,
        #     b_unit_domain_configuration=self.b_unit_domain_configuration)

        # # TODO: added here to avoid crashing because missing column
        # if self.b_unit_domain_configuration.b_unit_bie_enum != BClearerNeo4jEntificationBunits.LOAD_NEO4J:
        #     register_in_parallel_bie_universe(
        #         bie_universe=self.neo4j_import_spoke_bclearer_run_universe.parallel_bie_universe,
        #         b_unit_domain_configuration=self.b_unit_domain_configuration)

    @abstractmethod
    def b_unit_process_function(
        self,
    ) -> ():
        pass



from bclearer_core.common_knowledge.digitialisation_level_stereotype_matched_ea_objects import (
    DigitalisationLevelStereotypeMatchedEaObjects,
)
from bclearer_core.nf.types.nf_column_types import (
    NfColumnTypes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.collection_types.nf_ea_com_collection_types import (
    NfEaComCollectionTypes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.column_types.nf_ea_com_column_types import (
    NfEaComColumnTypes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.nf_ea_com_universes import (
    NfEaComUniverses,
)
from pandas import DataFrame


def get_digitalisation_level_stereotypes_usages(
    nf_ea_com_universe: NfEaComUniverses,
) -> DataFrame:
    digitalisation_level_stereotype_ea_guids = (
        DigitalisationLevelStereotypeMatchedEaObjects.get_ea_guids()
    )

    ea_stereotypes = (
        nf_ea_com_universe.nf_ea_com_registry.get_ea_stereotypes()
    )

    digitalisation_level_stereotypes = ea_stereotypes[
        ea_stereotypes[
            NfEaComColumnTypes.EXPLICIT_OBJECTS_EA_GUID.column_name
        ].isin(
            digitalisation_level_stereotype_ea_guids,
        )
    ]

    digitalisation_level_stereotypes_nf_uuids = set(
        digitalisation_level_stereotypes[
            NfColumnTypes.NF_UUIDS.column_name
        ],
    )

    ea_stereotype_usages = nf_ea_com_universe.nf_ea_com_registry.dictionary_of_collections[
        NfEaComCollectionTypes.STEREOTYPE_USAGE
    ]

    digitalisation_level_stereotypes_usages = ea_stereotype_usages[
        ea_stereotype_usages[
            "stereotype_nf_uuids"
        ].isin(
            digitalisation_level_stereotypes_nf_uuids,
        )
    ]

    return digitalisation_level_stereotypes_usages



from bclearer_core.common_knowledge.matched_objects import (
    MatchedEaObjects,
)


class BespokeNameToInstanceConfigurationObjects:
    def __init__(
        self,
        matched_naming_space_type: MatchedEaObjects,
        name_instance_attribute_name: str = None,
        matched_name_instance_type: MatchedEaObjects = None,
        package_name: str = None,
    ):
        self.matched_naming_space_type = (
            matched_naming_space_type
        )

        self.name_instance_attribute_name = (
            name_instance_attribute_name
        )

        self.matched_name_instance_type = (
            matched_name_instance_type
        )

        self.package_name = package_name

    def __enter__(self):
        return self

    def __exit__(
        self,
        exception_type,
        exception_value,
        traceback,
    ):
        pass



"""Template for pipeline runner."""

PIPELINE_RUNNER_TEMPLATE = """from bclearer_orchestration_services.reporting_service.wrappers.run_and_log_function_wrapper_latest import (
    run_and_log_function,
)
from bclearer_pipelines.{domain_name}.b_source.{pipeline_name}.orchestrators.pipeline.{pipeline_name}_orchestrator import (
    orchestrate_{pipeline_name},
)


@run_and_log_function()
def run_{pipeline_name}() -> None:
    orchestrate_{pipeline_name}()
"""



from enum import Enum, auto


class LoggingInspectionLevelBEnums(
    Enum
):
    NOT_SET = auto()

    VERBOSE = auto()

    INFO = auto()

    WARNING = auto()

    ERROR = auto()

    CRITICAL = auto()

    def __enum_level_weight(
        self,
    ) -> str:
        enum_level_weight = (
            enum_level_weight_mapping[
                self
            ]
        )

        return enum_level_weight

    enum_level_weight = property(
        fget=__enum_level_weight
    )


enum_level_weight_mapping = {
    LoggingInspectionLevelBEnums.NOT_SET: int(),
    LoggingInspectionLevelBEnums.VERBOSE: 1,
    LoggingInspectionLevelBEnums.INFO: 2,
    LoggingInspectionLevelBEnums.WARNING: 3,
    LoggingInspectionLevelBEnums.ERROR: 4,
    LoggingInspectionLevelBEnums.CRITICAL: 5,
}



from bclearer_core.nf.python_extensions.collections.nf_dictionary_try_get_results import (
    NfDictionaryTryGetResults,
)


class NfDictionaries(dict):
    def __init__(self):
        dict.__init__(self)

    def try_get_value(
        self,
        key,
    ) -> NfDictionaryTryGetResults:
        nf_dictionary_try_get_result = (
            NfDictionaryTryGetResults()
        )

        nf_dictionary_try_get_result.key_exists = (
            key in self.keys()
        )

        if (
            nf_dictionary_try_get_result.key_exists
        ):
            nf_dictionary_try_get_result.value = self.get(
                key,
            )
        else:
            nf_dictionary_try_get_result.value = (
                None
            )

        return (
            nf_dictionary_try_get_result
        )



from enum import auto, unique

from bclearer_core.common_knowledge.operation_types import (
    OperationTypes,
)


@unique
class UniverseModificationOperationTypes(
    OperationTypes,
):
    ADD_DEPENDENCY_TO_INSTANCES_OF_TYPE = (
        auto()
    )

    def __operation_name(self) -> str:
        operation_name = (
            operation_name_mapping[self]
        )

        return operation_name

    operation_name = property(
        fget=__operation_name,
    )


operation_name_mapping = {
    UniverseModificationOperationTypes.ADD_DEPENDENCY_TO_INSTANCES_OF_TYPE: "add_dependency_to_instances_of_type",
}



from bclearer_core.nf.python_extensions.collections.nf_dictionaries import (
    NfDictionaries,
)


class NfBimappings:
    def __init__(self, map: dict):
        NfBimappings.__populate_internal_dictionaries(
            self,
            map=map,
        )

    def try_get_range_using_domain(
        self,
        domain_key,
    ):
        range_value = self.__range_keyed_on_domain.try_get_value(
            key=domain_key,
        )

        return range_value

    def try_get_domain_using_range(
        self,
        range_key,
    ):
        domain_value = self.__domain_keyed_on_range.try_get_value(
            key=range_key,
        )

        return domain_value

    def get_range(self):
        range = (
            self.__domain_keyed_on_range.keys()
        )

        return range

    def get_domain(self):
        domain = (
            self.__range_keyed_on_domain.keys()
        )

        return domain

    def get_range_keyed_on_domain(self):
        range_keyed_on_domain = (
            self.__range_keyed_on_domain
        )

        return range_keyed_on_domain

    def get_domain_keyed_on_range(self):
        domain_keyed_on_range = (
            self.__domain_keyed_on_range
        )

        return domain_keyed_on_range

    def __populate_internal_dictionaries(
        self,
        map: dict,
    ):
        self.__domain_keyed_on_range = (
            NfDictionaries()
        )

        self.__range_keyed_on_domain = (
            NfDictionaries()
        )

        for (
            domain_value,
            range_value,
        ) in map.items():
            self.add_mapping(
                domain_value=domain_value,
                range_value=range_value,
            )

    def add_mapping(
        self,
        domain_value,
        range_value,
    ):
        self.__domain_keyed_on_range[
            range_value
        ] = domain_value

        self.__range_keyed_on_domain[
            domain_value
        ] = range_value

    @staticmethod
    def __populate_range_keyed_on_range(
        map: NfDictionaries,
    ) -> NfDictionaries:
        domain_keyed_on_range = (
            NfDictionaries()
        )

        for (
            domain_value,
            range_value,
        ) in map:
            domain_keyed_on_range[
                range_value
            ] = domain_value

        return domain_keyed_on_range



from bclearer_core.common_knowledge.universe_modification_operation_types import (
    UniverseModificationOperationTypes,
)
from bclearer_core.configurations.add_dependency_to_instances_of_type_configuration_objects import (
    AddDependencyToInstancesOfTypeConfigurationObjects,
)
from bclearer_core.configurations.universe_modification_operation_configurations import (
    UniverseModificationOperationConfigurations,
)
from bclearer_core.substages.operations.b_evolve.universe_modification_operations.universe_modifiers.dependency_to_instances_of_type_adder import (
    add_dependency_to_instances_of_type,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.nf_ea_com_universes import (
    NfEaComUniverses,
)
from bclearer_interop_services.ea_interop_service.session.orchestrators.ea_tools_session_managers import (
    EaToolsSessionManagers,
)
from bclearer_interop_services.ea_interop_service.session.processes.creators.empty_nf_ea_com_universe_creator import (
    create_empty_nf_ea_universe,
)


class UniverseModificationOperationsSubstages:
    def __init__(
        self,
        ea_tools_session_manager: EaToolsSessionManagers,
        universe_modification_operation_configuration: UniverseModificationOperationConfigurations,
        content_universe: NfEaComUniverses,
    ):
        self.ea_tools_session_manager = (
            ea_tools_session_manager
        )

        self.universe_modification_operation_configuration = universe_modification_operation_configuration

        self.content_universe = (
            content_universe
        )

    def __enter__(self):
        return self

    def __exit__(
        self,
        exception_type,
        exception_value,
        traceback,
    ):
        pass

    def run(self) -> NfEaComUniverses:
        output_universe = create_empty_nf_ea_universe(
            ea_tools_session_manager=self.ea_tools_session_manager,
            short_name=self.universe_modification_operation_configuration.output_universe_short_name,
        )

        universe_modification_configuration_object = (
            self.universe_modification_operation_configuration.universe_modification_configuration_object
        )

        if (
            self.universe_modification_operation_configuration.operation_type
            == UniverseModificationOperationTypes.ADD_DEPENDENCY_TO_INSTANCES_OF_TYPE
        ):
            if not isinstance(
                universe_modification_configuration_object,
                AddDependencyToInstancesOfTypeConfigurationObjects,
            ):
                raise TypeError

            add_dependency_to_instances_of_type(
                content_universe=self.content_universe,
                output_universe=output_universe,
                matched_target_object=universe_modification_configuration_object.matched_target_type,
                matched_source_objects_type=universe_modification_configuration_object.matched_source_objects_type,
            )
        else:
            raise NotImplementedError

        return output_universe



from enum import auto, unique

from bclearer_core.common_knowledge.operation_types import (
    OperationTypes,
)


@unique
class ContentOperationTypes(
    OperationTypes,
):
    MERGE_UNIVERSES = auto()

    def __operation_name(self) -> str:
        operation_name = (
            operation_name_mapping[self]
        )

        return operation_name

    operation_name = property(
        fget=__operation_name,
    )


operation_name_mapping = {
    ContentOperationTypes.MERGE_UNIVERSES: "merge_universes",
}



import json
import os
import shutil
import subprocess
import sys
from io import StringIO
from pathlib import Path
from unittest import mock

import pytest
from bclearer_core.pipeline_builder.cli import (
    create_parser,
    interactive_config,
    run_cli,
    save_sample_config,
)


class TestPipelineBuilderCLI:
    """Test class for the pipeline builder CLI functionality."""

    @pytest.fixture(autouse=True)
    def setup_method(
        self,
        pipeline_output_folder_absolute_path,
    ):
        """Setup method that runs before each test."""
        # Clean up any existing test pipelines from previous runs
        self.test_domain_path = os.path.join(
            pipeline_output_folder_absolute_path,
            "pipelines",
            "test_domain",
        )

        if os.path.exists(
            self.test_domain_path
        ):
            shutil.rmtree(
                self.test_domain_path
            )

        # Create a temporary config file for CLI tests
        self.temp_config_path = os.path.join(
            pipeline_output_folder_absolute_path,
            "temp_config.json",
        )
        self.temp_sample_path = os.path.join(
            pipeline_output_folder_absolute_path,
            "sample_config.json",
        )

    def teardown_method(self):
        """Cleanup after tests."""
        if os.path.exists(
            self.temp_config_path
        ):
            os.remove(
                self.temp_config_path
            )

        if os.path.exists(
            self.temp_sample_path
        ):
            os.remove(
                self.temp_sample_path
            )

    def test_parser_create_command(
        self,
    ):
        """Test the CLI argument parser with create command."""
        parser = create_parser()
        args = parser.parse_args(
            [
                "create",
                "--config",
                "test.json",
                "--output",
                "/test/path",
            ]
        )

        assert args.command == "create"
        assert (
            args.config == "test.json"
        )
        assert (
            args.output == "/test/path"
        )
        assert args.interactive == False

    def test_parser_sample_command(
        self,
    ):
        """Test the CLI argument parser with sample command."""
        parser = create_parser()
        args = parser.parse_args(
            [
                "sample",
                "--output",
                "test_sample.json",
            ]
        )

        assert args.command == "sample"
        assert (
            args.output
            == "test_sample.json"
        )

    def test_save_sample_config(self):
        """Test saving a sample configuration file."""
        save_sample_config(
            self.temp_sample_path
        )

        # Check the file exists
        assert os.path.exists(
            self.temp_sample_path
        )

        # Check the content is valid JSON
        with open(
            self.temp_sample_path, "r"
        ) as f:
            config = json.load(f)

        # Verify basic structure
        assert "domain_name" in config
        assert "pipelines" in config
        assert (
            len(config["pipelines"]) > 0
        )

    @mock.patch("builtins.input")
    def test_interactive_config(
        self, mock_input
    ):
        """Test the interactive configuration process."""
        # Setup mock inputs
        mock_inputs = [
            "test_interactive_domain",  # Domain name
            "1",  # Number of pipelines
            "test_interactive_pipeline",  # Pipeline name
            "1",  # Number of thin slices
            "test_thin_slice",  # Thin slice name
            "y",  # Include Collect stage
            "y",  # Has direct b_units
            "ca,cb",  # B-unit names for collect
            "n",  # No sub-stages for collect
            "y",  # Include Load stage
            "y",  # Has direct b_units
            "la,lb",  # B-unit names for load
            "n",  # No sub-stages for load
            "y",  # Include Evolve stage
            "n",  # No direct b_units
            "y",  # Has sub-stages
            "2",  # Number of sub-stages
            "sub_stage_1",  # Sub-stage 1 name
            "ea1,ea2",  # B-unit names for sub-stage 1
            "sub_stage_2",  # Sub-stage 2 name
            "eb1,eb2",  # B-unit names for sub-stage 2
            "n",  # Don't include Assimilate stage
            "n",  # Don't include Reuse stage
        ]

        mock_input.side_effect = (
            mock_inputs
        )

        # Call the interactive config function
        config = interactive_config()

        # Check the configuration structure
        assert (
            config["domain_name"]
            == "test_interactive_domain"
        )
        assert (
            len(config["pipelines"])
            == 1
        )
        assert (
            config["pipelines"][0][
                "name"
            ]
            == "test_interactive_pipeline"
        )
        assert (
            len(
                config["pipelines"][0][
                    "thin_slices"
                ]
            )
            == 1
        )

        # Check stages
        stages = config["pipelines"][0][
            "thin_slices"
        ][0]["stages"]
        assert (
            len(stages) == 3
        )  # Only included Collect, Load, and Evolve

        # Check Collect stage
        collect_stage = next(
            s
            for s in stages
            if s["name"] == "1c_collect"
        )
        assert (
            len(
                collect_stage["b_units"]
            )
            == 2
        )
        assert collect_stage[
            "b_units"
        ] == ["ca", "cb"]

        # Check Evolve stage
        evolve_stage = next(
            s
            for s in stages
            if s["name"] == "3e_evolve"
        )
        assert (
            len(
                evolve_stage[
                    "sub_stages"
                ]
            )
            == 2
        )
        assert (
            evolve_stage["sub_stages"][
                0
            ]["name"]
            == "sub_stage_1"
        )
        assert evolve_stage[
            "sub_stages"
        ][0]["b_units"] == [
            "ea1",
            "ea2",
        ]

    @mock.patch("sys.argv")
    def test_cli_sample_command(
        self,
        mock_argv,
        pipeline_output_folder_absolute_path,
    ):
        """Test the CLI sample command."""
        # Setup mock command line args
        output_path = os.path.join(
            pipeline_output_folder_absolute_path,
            "cli_sample.json",
        )
        mock_argv.__getitem__.side_effect = lambda i: [
            "bclearer-pipeline-builder",
            "sample",
            "--output",
            output_path,
        ][
            i
        ]

        # Call the CLI function with mocked arguments
        with mock.patch(
            "sys.argv",
            [
                "bclearer-pipeline-builder",
                "sample",
                "--output",
                output_path,
            ],
        ):
            run_cli()

        # Check the sample file was created
        assert os.path.exists(
            output_path
        )

        # Check it contains valid JSON
        with open(
            output_path, "r"
        ) as f:
            config = json.load(f)

        # Verify basic structure
        assert "domain_name" in config
        assert "pipelines" in config
        assert (
            len(config["pipelines"]) > 0
        )



from bclearer_core.common_knowledge.digitialisation_level_stereotype_matched_ea_objects import (
    DigitalisationLevelStereotypeMatchedEaObjects,
)
from bclearer_core.nf.types.nf_column_types import (
    NfColumnTypes,
)
from bclearer_core.substages.operations.common.new_ea_objects_dictionary_creator import (
    create_new_ea_objects_dictionary,
)
from bclearer_core.substages.operations.common.new_stereotype_to_dictionary_adder import (
    add_new_stereotype_to_dictionary,
)
from bclearer_core.substages.operations.common.nf_ea_com_universe_updater import (
    update_nf_ea_com_universe_with_dictionary,
)
from bclearer_core.substages.operations.common.nf_uuid_from_ea_guid_from_collection_getter import (
    get_nf_uuid_from_ea_guid_from_collection,
)
from bclearer_core.substages.operations.common.stereotype_adder import (
    add_new_stereotype_usage_to_dictionary,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.collection_types.nf_ea_com_collection_types import (
    NfEaComCollectionTypes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.column_types.nf_ea_com_column_types import (
    NfEaComColumnTypes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.nf_ea_com_universes import (
    NfEaComUniverses,
)


def add_default_digitalisation_level_stereotype(
    nf_ea_com_universe: NfEaComUniverses,
    default_digitalisation_level_stereotype: DigitalisationLevelStereotypeMatchedEaObjects,
) -> None:
    __add_stereotypes_if_required(
        nf_ea_com_universe=nf_ea_com_universe,
    )

    __add_stereotype_usage(
        nf_ea_com_universe=nf_ea_com_universe,
        default_digitalisation_level_stereotype=default_digitalisation_level_stereotype,
    )


def __add_stereotypes_if_required(
    nf_ea_com_universe: NfEaComUniverses,
) -> None:
    new_ea_objects_dictionary = (
        create_new_ea_objects_dictionary()
    )

    ea_stereotypes = (
        nf_ea_com_universe.nf_ea_com_registry.get_ea_stereotypes()
    )

    ea_stereotype_ea_guids = set(
        ea_stereotypes[
            NfEaComColumnTypes.EXPLICIT_OBJECTS_EA_GUID.column_name
        ],
    )

    for digitalisation_level_stereotype_matched_ea_object in DigitalisationLevelStereotypeMatchedEaObjects:
        __add_stereotype_if_required(
            new_ea_objects_dictionary=new_ea_objects_dictionary,
            digitalisation_level_stereotype_matched_ea_object=digitalisation_level_stereotype_matched_ea_object,
            ea_stereotype_ea_guids=ea_stereotype_ea_guids,
        )

    update_nf_ea_com_universe_with_dictionary(
        nf_ea_com_universe=nf_ea_com_universe,
        new_ea_objects_dictionary=new_ea_objects_dictionary,
    )


def __add_stereotype_if_required(
    new_ea_objects_dictionary: dict,
    digitalisation_level_stereotype_matched_ea_object: DigitalisationLevelStereotypeMatchedEaObjects,
    ea_stereotype_ea_guids: set,
) -> None:
    if (
        digitalisation_level_stereotype_matched_ea_object.ea_guid
        in ea_stereotype_ea_guids
    ):
        return

    add_new_stereotype_to_dictionary(
        new_stereotype_dictionary=new_ea_objects_dictionary[
            NfEaComCollectionTypes.EA_STEREOTYPES
        ],
        stereotype_ea_guid=digitalisation_level_stereotype_matched_ea_object.ea_guid,
        stereotype_name=digitalisation_level_stereotype_matched_ea_object.object_name,
        stereotype_applies_to="class",
        stereotype_style=digitalisation_level_stereotype_matched_ea_object.style,
    )


def __add_stereotype_usage(
    nf_ea_com_universe: NfEaComUniverses,
    default_digitalisation_level_stereotype: DigitalisationLevelStereotypeMatchedEaObjects,
) -> None:
    new_ea_objects_dictionary = (
        create_new_ea_objects_dictionary()
    )

    default_digitalisation_level_stereotype_nf_uuid = get_nf_uuid_from_ea_guid_from_collection(
        nf_ea_com_universe=nf_ea_com_universe,
        collection_type=NfEaComCollectionTypes.EA_STEREOTYPES,
        ea_guid=default_digitalisation_level_stereotype.ea_guid,
    )

    ea_classifiers = (
        nf_ea_com_universe.nf_ea_com_registry.get_ea_classifiers()
    )

    ea_classifier_nf_uuids = set(
        ea_classifiers[
            NfColumnTypes.NF_UUIDS.column_name
        ],
    )

    for (
        ea_classifier_nf_uuid
    ) in ea_classifier_nf_uuids:
        add_new_stereotype_usage_to_dictionary(
            new_stereotype_usage_dictionary=new_ea_objects_dictionary[
                NfEaComCollectionTypes.STEREOTYPE_USAGE
            ],
            client_nf_uuid=ea_classifier_nf_uuid,
            client_collection_type=NfEaComCollectionTypes.EA_CLASSIFIERS,
            stereotype_nf_uuid=default_digitalisation_level_stereotype_nf_uuid,
        )

    update_nf_ea_com_universe_with_dictionary(
        nf_ea_com_universe=nf_ea_com_universe,
        new_ea_objects_dictionary=new_ea_objects_dictionary,
    )



from bclearer_core.common_knowledge.bclearer_matched_ea_objects import (
    BclearerMatchedEaObjects,
)
from bclearer_core.common_knowledge.matched_objects import (
    MatchedEaObjects,
)
from bclearer_core.nf.types.nf_column_types import (
    NfColumnTypes,
)
from bclearer_core.substages.operations.common.connector_adder import (
    add_new_connector_to_dictionary,
)
from bclearer_core.substages.operations.common.instances_nf_uuids_getter import (
    get_instances_nf_uuids_of_matched_type,
    get_instances_nf_uuids_of_type_nf_uuid,
)
from bclearer_core.substages.operations.common.new_ea_objects_dictionary_creator import (
    create_new_ea_objects_dictionary,
)
from bclearer_core.substages.operations.common.nf_ea_com_universe_updater import (
    update_nf_ea_com_universe_with_dictionary,
)
from bclearer_core.substages.operations.common.nf_uuid_from_ea_guid_from_collection_getter import (
    get_nf_uuid_from_ea_guid_from_collection,
)
from bclearer_core.substages.operations.common.stereotype_adder import (
    add_new_stereotype_usage_to_dictionary,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.collection_types.nf_ea_com_collection_types import (
    NfEaComCollectionTypes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.column_types.nf_ea_com_column_types import (
    NfEaComColumnTypes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.nf_ea_com_universes import (
    NfEaComUniverses,
)
from bclearer_interop_services.ea_interop_service.nf_ea_common.common_knowledge.ea_connector_types import (
    EaConnectorTypes,
)
from bclearer_orchestration_services.reporting_service.wrappers.run_and_log_function_wrapper import (
    run_and_log_function,
)


@run_and_log_function
def annotate_names_and_instances(
    nf_ea_com_universe: NfEaComUniverses,
    matched_name_instance_type: MatchedEaObjects,
):
    new_ea_objects_dictionary = (
        create_new_ea_objects_dictionary()
    )

    name_instance_instances = get_instances_nf_uuids_of_matched_type(
        nf_ea_com_universe=nf_ea_com_universe,
        matched_type=matched_name_instance_type,
    )

    for (
        name_instance_instance_nf_uuid
    ) in name_instance_instances:
        __annotate_naming_space_instances(
            nf_ea_com_universe=nf_ea_com_universe,
            name_instance_instance_nf_uuid=name_instance_instance_nf_uuid,
            new_ea_objects_dictionary=new_ea_objects_dictionary,
        )

    update_nf_ea_com_universe_with_dictionary(
        nf_ea_com_universe=nf_ea_com_universe,
        new_ea_objects_dictionary=new_ea_objects_dictionary,
    )


@run_and_log_function
def __annotate_naming_space_instances(
    nf_ea_com_universe: NfEaComUniverses,
    name_instance_instance_nf_uuid: str,
    new_ea_objects_dictionary: dict,
):
    name_instances_nf_uuid = get_nf_uuid_from_ea_guid_from_collection(
        nf_ea_com_universe=nf_ea_com_universe,
        collection_type=NfEaComCollectionTypes.EA_CLASSIFIERS,
        ea_guid=BclearerMatchedEaObjects.NAME_INSTANCES.ea_guid,
    )

    add_new_connector_to_dictionary(
        new_connector_dictionary=new_ea_objects_dictionary[
            NfEaComCollectionTypes.EA_CONNECTORS
        ],
        place_1_nf_uuid=name_instance_instance_nf_uuid,
        place_2_nf_uuid=name_instances_nf_uuid,
        connector_type=EaConnectorTypes.GENERALIZATION,
    )

    __add_stereotype_usages(
        nf_ea_com_universe=nf_ea_com_universe,
        name_instance_instance_nf_uuid=name_instance_instance_nf_uuid,
        new_ea_objects_dictionary=new_ea_objects_dictionary,
    )


def __add_stereotype_usages(
    nf_ea_com_universe: NfEaComUniverses,
    name_instance_instance_nf_uuid: str,
    new_ea_objects_dictionary: dict,
):
    name_instance_instances = get_instances_nf_uuids_of_type_nf_uuid(
        nf_ea_com_universe=nf_ea_com_universe,
        nf_uuid=name_instance_instance_nf_uuid,
    )

    name_types_instances_stereotype_nf_uuid = get_nf_uuid_from_ea_guid_from_collection(
        nf_ea_com_universe=nf_ea_com_universe,
        collection_type=NfEaComCollectionTypes.EA_STEREOTYPES,
        ea_guid=BclearerMatchedEaObjects.NAME_TYPES_INSTANCES_STEREOTYPE.ea_guid,
    )

    ea_connectors = (
        nf_ea_com_universe.nf_ea_com_registry.get_ea_connectors()
    )

    ea_associations = ea_connectors[
        ea_connectors[
            NfEaComColumnTypes.CONNECTORS_ELEMENT_TYPE_NAME.column_name
        ]
        == EaConnectorTypes.ASSOCIATION.type_name
    ]

    filtered_associations = ea_associations[
        ea_associations[
            NfEaComColumnTypes.ELEMENTS_SUPPLIER_PLACE1_END_CONNECTORS.column_name
        ].isin(name_instance_instances)
    ]

    association_nf_uuids = set(
        filtered_associations[
            NfColumnTypes.NF_UUIDS.column_name
        ],
    )

    for (
        association_nf_uuid
    ) in association_nf_uuids:
        add_new_stereotype_usage_to_dictionary(
            new_stereotype_usage_dictionary=new_ea_objects_dictionary[
                NfEaComCollectionTypes.STEREOTYPE_USAGE
            ],
            client_nf_uuid=association_nf_uuid,
            client_collection_type=NfEaComCollectionTypes.EA_CONNECTORS,
            stereotype_nf_uuid=name_types_instances_stereotype_nf_uuid,
        )



from bclearer_core.common_knowledge.attribute_to_associations_operation_subtypes import (
    AttributeToAssociationOperationSubtypes,
)
from bclearer_core.substages.operations.b_evolve.adjustment_operations.convert_attributes_to_associations.attributes_converter import (
    convert_attributes,
)
from bclearer_core.substages.operations.b_evolve.adjustment_operations.convert_attributes_to_associations.attributes_to_convert_getter import (
    get_attributes_to_convert,
)
from bclearer_core.substages.operations.b_evolve.adjustment_operations.remove_attributes.attribute_remover import (
    remove_attributes,
)
from bclearer_core.substages.operations.b_evolve.common.new_root_package_creator import (
    create_root_package,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.nf_ea_com_universes import (
    NfEaComUniverses,
)
from bclearer_interop_services.ea_interop_service.nf_ea_common.common_knowledge.ea_association_direction_types import (
    EaAssociationDirectionTypes,
)
from bclearer_orchestration_services.reporting_service.reporters.log_with_datetime import (
    log_message,
)


def convert_attributes_to_associations(
    content_universe: NfEaComUniverses,
    adjustment_universe: NfEaComUniverses,
    output_universe: NfEaComUniverses,
    direction: EaAssociationDirectionTypes,
    package_name: str,
    # attribute_to_association_operation_subtype: AttributeToAssociationOperationSubtypes,
) -> None:
    __run_input_checks()

    __run_operation(
        content_universe=content_universe,
        adjustment_universe=adjustment_universe,
        output_universe=output_universe,
        direction=direction,
        package_name=package_name,
        # attribute_to_association_operation_subtype=attribute_to_association_operation_subtype,
    )


def __run_input_checks():
    pass


def __run_operation(
    content_universe: NfEaComUniverses,
    adjustment_universe: NfEaComUniverses,
    output_universe: NfEaComUniverses,
    direction: EaAssociationDirectionTypes,
    package_name: str,
    # attribute_to_association_operation_subtype: AttributeToAssociationOperationSubtypes,
):
    log_message(
        message="ADJUSTMENT OPERATION: Convert attributes - started",
    )

    attributes_to_convert = get_attributes_to_convert(
        content_universe=content_universe,
        adjustment_universe=adjustment_universe,
    )

    package_nf_uuid = create_root_package(
        nf_ea_com_universe=content_universe,
        package_name=package_name,
    )

    __copy_collections(
        content_universe=content_universe,
        output_universe=output_universe,
    )

    convert_attributes(
        output_universe=output_universe,
        attributes_to_convert=attributes_to_convert,
        direction=direction,
        package_nf_uuid=package_nf_uuid,
        # attribute_to_association_operation_subtype=attribute_to_association_operation_subtype,
    )

    log_message(
        message="ADJUSTMENT OPERATION: Convert attributes - removing converted attributes",
    )

    remove_attributes(
        content_universe=output_universe,
        adjustment_universe=adjustment_universe,
        output_universe=output_universe,
    )

    log_message(
        message="ADJUSTMENT OPERATION: Convert attributes - finished",
    )


def __copy_collections(
    content_universe: NfEaComUniverses,
    output_universe: NfEaComUniverses,
) -> None:
    content_collections_dictionary = (
        content_universe.nf_ea_com_registry.dictionary_of_collections
    )

    output_collections_dictionary = (
        output_universe.nf_ea_com_registry.dictionary_of_collections
    )

    for (
        content_collection_type,
        content_collection_table,
    ) in (
        content_collections_dictionary.items()
    ):
        output_collections_dictionary[
            content_collection_type
        ] = content_collection_table



from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.collection_types.nf_ea_com_collection_types import (
    NfEaComCollectionTypes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.column_types.nf_ea_com_column_types import (
    NfEaComColumnTypes,
)
from bclearer_interop_services.ea_interop_service.nf_ea_common.common_knowledge.ea_property_types import (
    EaPropertyTypes,
)
from bclearer_orchestration_services.identification_services.uuid_service.uuid_helpers.uuid_factory import (
    create_new_uuid,
)


def add_new_stereotype_usage_to_dictionary(
    new_stereotype_usage_dictionary: dict,
    client_nf_uuid: str,
    client_collection_type: NfEaComCollectionTypes,
    stereotype_nf_uuid: str,
) -> None:
    if (
        client_collection_type
        == NfEaComCollectionTypes.EA_CLASSIFIERS
    ):
        property_type = (
            EaPropertyTypes.ELEMENT_PROPERTY.type_name
        )

    elif (
        client_collection_type
        == NfEaComCollectionTypes.EA_CONNECTORS
    ):
        property_type = (
            EaPropertyTypes.CONNECTOR_PROPERTY.type_name
        )

    elif (
        client_collection_type
        == NfEaComCollectionTypes.EA_ATTRIBUTES
    ):
        property_type = (
            EaPropertyTypes.ATTRIBUTE_PROPERTY.type_name
        )

    else:
        raise NotImplementedError

    stereotype_usage_row_dictionary = {
        NfEaComColumnTypes.STEREOTYPE_CLIENT_NF_UUIDS.column_name: client_nf_uuid,
        "stereotype_nf_uuids": stereotype_nf_uuid,
        NfEaComColumnTypes.STEREOTYPE_PROPERTY_TYPE.column_name: property_type,
    }

    new_stereotype_usage_dictionary[
        create_new_uuid()
    ] = stereotype_usage_row_dictionary



import json


def load_configuration(json_file_path):
    with open(json_file_path) as file:
        config = json.load(file)
        pretty_config = json.dumps(
            config,
            indent=4,
        )
        print(
            "loaded configuration:\n",
            pretty_config,
        )
    return config



from bclearer_core.common_knowledge.convention_shift_operation_types import (
    ConventionShiftOperationTypes,
)
from bclearer_core.configurations.operation_configurations import (
    OperationConfigurations,
)


class GeneraliseNamesConfigurations(
    OperationConfigurations,
):
    def __init__(
        self,
        convention_shift_operation_type: ConventionShiftOperationTypes,
        list_of_configuration_objects: list,
        output_universe_short_name: str,
    ):
        super().__init__(
            operation_type=convention_shift_operation_type,
        )

        self.list_of_configuration_objects = list_of_configuration_objects

        self.output_universe_short_name = (
            output_universe_short_name
        )

    def __enter__(self):
        return self

    def __exit__(
        self,
        exception_type,
        exception_value,
        traceback,
    ):
        pass



import json
import os
import shutil
from pathlib import Path

import pytest
from bclearer_core.pipeline_builder.generator import (
    generate_pipeline,
    update_pipeline,
)


class TestPipelineUpdate:
    """Test class for the pipeline update functionality."""

    @pytest.fixture(autouse=True)
    def setup_method(
        self,
        pipeline_output_folder_absolute_path,
        pipeline_builder_configuration_file_path,
    ):
        """Setup method that runs before each test."""
        # Load the configuration to get domain_name
        with open(
            pipeline_builder_configuration_file_path,
            "r",
        ) as f:
            self.config_dict = (
                json.load(f)
            )

        # Get the domain name from the config
        domain_name = self.config_dict[
            "domain_name"
        ]

        # Clean up any existing test pipelines from previous runs
        self.domain_path = os.path.join(
            pipeline_output_folder_absolute_path,
            f"{domain_name}_pipelines",
        )

        if os.path.exists(
            self.domain_path
        ):
            shutil.rmtree(
                self.domain_path
            )

        # Store paths for use in tests
        self.output_path = pipeline_output_folder_absolute_path
        self.pipeline_builder_config_path = pipeline_builder_configuration_file_path

    def test_update_pipeline_new_b_unit(
        self,
    ):
        """Test updating a pipeline by adding a new b_unit."""
        # First generate the initial pipeline
        generate_pipeline(
            self.config_dict,
            self.output_path,
        )

        # Make sure the initial pipeline exists
        domain_name = self.config_dict[
            "domain_name"
        ]
        first_pipeline = (
            self.config_dict[
                "pipelines"
            ][0]["name"]
        )
        first_stage = self.config_dict[
            "pipelines"
        ][0]["thin_slices"][0][
            "stages"
        ][
            0
        ][
            "name"
        ]

        # Create updated config with a new b_unit in the first stage
        updated_config = (
            self.config_dict.copy()
        )
        updated_config["pipelines"][0][
            "thin_slices"
        ][0]["stages"][0][
            "b_units"
        ].append(
            "new_b_unit"
        )

        # Update the pipeline
        updated_path = update_pipeline(
            updated_config,
            self.domain_path,
        )

        # Check that the new b_unit file was created
        new_b_unit_path = os.path.join(
            self.domain_path,
            "b_source",
            first_pipeline,
            "objects",
            "b_units",
            f"{first_pipeline}_{first_stage}",
            "new_b_unit_b_units.py",
        )

        assert os.path.exists(
            new_b_unit_path
        )

        # Check content of new b_unit file
        with open(
            new_b_unit_path, "r"
        ) as f:
            content = f.read()
            assert (
                "class NewBUnitBUnits:"
                in content
            )
            assert (
                "def run(self) -> None:"
                in content
            )

    def test_update_pipeline_new_stage(
        self,
    ):
        """Test updating a pipeline by adding a new stage."""
        # First generate the initial pipeline
        generate_pipeline(
            self.config_dict,
            self.output_path,
        )

        # Create updated config with a new stage
        updated_config = (
            self.config_dict.copy()
        )
        updated_config["pipelines"][0][
            "thin_slices"
        ][0]["stages"].append(
            {
                "name": "6x_extend",
                "sub_stages": [],
                "b_units": [
                    "new_extension_b_unit"
                ],
            }
        )

        # Update the pipeline
        updated_path = update_pipeline(
            updated_config,
            self.domain_path,
        )

        # Check that the new stage orchestrator was created
        domain_name = self.config_dict[
            "domain_name"
        ]
        first_pipeline = (
            self.config_dict[
                "pipelines"
            ][0]["name"]
        )

        new_stage_orchestrator_path = os.path.join(
            self.domain_path,
            "b_source",
            first_pipeline,
            "orchestrators",
            "stages",
            f"{first_pipeline}_6x_extend_orchestrator.py",
        )

        assert os.path.exists(
            new_stage_orchestrator_path
        )

        # Check that the new b_unit was created
        new_b_unit_path = os.path.join(
            self.domain_path,
            "b_source",
            first_pipeline,
            "objects",
            "b_units",
            f"{first_pipeline}_6x_extend",
            "new_extension_b_unit_b_units.py",
        )

        assert os.path.exists(
            new_b_unit_path
        )

        # Check that the thin slice orchestrator was updated to include the new stage
        thin_slice_name = (
            self.config_dict[
                "pipelines"
            ][0]["thin_slices"][0][
                "name"
            ]
        )
        thin_slice_orchestrator_path = os.path.join(
            self.domain_path,
            "b_source",
            first_pipeline,
            "orchestrators",
            "thin_slices",
            f"{thin_slice_name}_orchestrator.py",
        )

        with open(
            thin_slice_orchestrator_path,
            "r",
        ) as f:
            content = f.read()
            assert (
                f"orchestrate_{first_pipeline}_6x_extend()"
                in content
            )

    def test_update_pipeline_new_pipeline(
        self,
    ):
        """Test updating by adding a completely new pipeline."""
        # First generate the initial pipeline
        generate_pipeline(
            self.config_dict,
            self.output_path,
        )

        # Create updated config with a new pipeline
        updated_config = (
            self.config_dict.copy()
        )
        updated_config[
            "pipelines"
        ].append(
            {
                "name": "new_pipeline",
                "thin_slices": [
                    {
                        "name": "new_slice",
                        "stages": [
                            {
                                "name": "1c_collect",
                                "sub_stages": [],
                                "b_units": [
                                    "collector_b_unit"
                                ],
                            }
                        ],
                    }
                ],
            }
        )

        # Update the pipeline
        update_pipeline(
            updated_config,
            self.domain_path,
        )

        # Check that the new pipeline directory was created
        domain_name = self.config_dict[
            "domain_name"
        ]
        new_pipeline_path = (
            os.path.join(
                self.domain_path,
                "b_source",
                "new_pipeline",
            )
        )
        assert os.path.exists(
            new_pipeline_path
        )

        # Check that the new pipeline runner was created
        new_pipeline_runner_path = os.path.join(
            self.domain_path,
            "b_source",
            "app_runners",
            "runners",
            "new_pipeline_runner.py",
        )
        assert os.path.exists(
            new_pipeline_runner_path
        )

        # Check that the pipelines runner was updated to include the new pipeline
        pipelines_runner_path = os.path.join(
            self.domain_path,
            "b_source",
            "app_runners",
            "runners",
            f"{domain_name}_b_clearer_pipelines_runner.py",
        )

        with open(
            pipelines_runner_path, "r"
        ) as f:
            content = f.read()
            assert (
                f"from {domain_name}.b_source.app_runners.runners.new_pipeline_runner import"
                in content
            )
            assert (
                "run_new_pipeline()"
                in content
            )

    def test_update_pipeline_domain_name_mismatch(
        self,
    ):
        """Test that updating with a mismatched domain name raises an error."""
        # First generate the initial pipeline
        generate_pipeline(
            self.config_dict,
            self.output_path,
        )

        # Create updated config with a different domain name
        updated_config = (
            self.config_dict.copy()
        )
        updated_config[
            "domain_name"
        ] = "different_domain"

        # Updating the pipeline should raise a ValueError
        with pytest.raises(
            ValueError
        ) as excinfo:
            update_pipeline(
                updated_config,
                self.domain_path,
            )

        assert (
            "Domain name in configuration"
            in str(excinfo.value)
        )
        assert (
            "does not match existing domain name"
            in str(excinfo.value)
        )



"""Tests for the template extractor."""

import os
import shutil
import tempfile
from pathlib import Path

import pytest
from bclearer_core.pipeline_builder.template_extractor import (
    TemplateExtractor,
    update_templates_from_pipeline,
)


class TestTemplateExtractor:
    """Test class for the template extractor functionality."""

    @pytest.fixture(scope="function")
    def temp_dir(self):
        """Create a temporary directory for testing."""
        temp_dir = tempfile.mkdtemp()
        yield temp_dir
        shutil.rmtree(temp_dir)

    def test_template_extractor_initialization_errors(
        self, temp_dir
    ):
        """Test error handling during initialization."""
        # Test non-existent template path
        with pytest.raises(
            FileNotFoundError
        ) as excinfo:
            TemplateExtractor(
                "/path/does/not/exist"
            )
        assert (
            "Template path not found"
            in str(excinfo.value)
        )

        # Test file as template path (not a directory)
        test_file = os.path.join(
            temp_dir, "test.txt"
        )
        with open(test_file, "w") as f:
            f.write("test")

        with pytest.raises(
            NotADirectoryError
        ) as excinfo:
            TemplateExtractor(test_file)
        assert (
            "Template path is not a directory"
            in str(excinfo.value)
        )

    def test_create_minimal_template(
        self,
        temp_dir,
        template_pipeline_folder_absolute_path,
    ):
        """Test creating and using a minimal template structure."""
        # Create a minimal template structure for testing
        template_dir = os.path.join(
            temp_dir, "test_template"
        )
        os.makedirs(template_dir)

        # Create b_source directory
        b_source_dir = os.path.join(
            template_dir, "b_source"
        )
        os.makedirs(b_source_dir)

        # Create app_runners directory
        app_runners_dir = os.path.join(
            b_source_dir, "app_runners"
        )
        os.makedirs(app_runners_dir)

        # Create a simple application runner file
        app_runner_file = os.path.join(
            app_runners_dir,
            "aa_b_clearer_pipeline_b_application_runner.py",
        )
        with open(
            app_runner_file, "w"
        ) as f:
            f.write(
                """from bclearer_orchestration_services.b_app_runner_service.b_application_runner import (
    run_b_application,
)
from aa.b_source.app_runners.runners.aa_b_clearer_pipelines_runner import (
    run_aa_b_clearer_pipelines,
)


def run_aa_b_clearer_pipeline_b_application() -> (
    None
):
    run_b_application(
        app_startup_method=run_aa_b_clearer_pipelines
    )
"""
            )

        # Create output directory for extracted templates
        output_dir = os.path.join(
            temp_dir, "templates"
        )

        # Extract templates
        extractor = TemplateExtractor(
            template_dir
        )
        extractor.extract_templates(
            output_dir
        )

        # Check that the application runner template was extracted
        app_runner_template_file = (
            os.path.join(
                output_dir,
                "application_runner.py",
            )
        )
        assert os.path.exists(
            app_runner_template_file
        )

        # Check template content was properly formatted
        with open(
            app_runner_template_file,
            "r",
        ) as f:
            content = f.read()
            assert (
                "APPLICATION_RUNNER_TEMPLATE"
                in content
            )
            assert (
                "{domain_name}"
                in content
            )  # Should have replaced "aa" with "{domain_name}"
            assert (
                "aa" not in content
            )  # Should not contain the original domain name

    def test_update_templates_from_pipeline(
        self,
        temp_dir,
        template_pipeline_folder_absolute_path,
    ):
        """Test the update_templates_from_pipeline function."""
        # Create output directory
        output_dir = os.path.join(
            temp_dir, "templates"
        )

        # Call the function
        update_templates_from_pipeline(
            template_pipeline_folder_absolute_path,
            output_dir,
        )

        # Check that key template files were created
        expected_files = [
            "application_runner.py",
            "b_unit.py",
            "b_unit_creator_and_runner.py",
            "conftest.py",
            "e2e_test.py",
            "pipeline_orchestrator.py",
            "pipeline_runner.py",
            "pipelines_runner.py",
            "stage_orchestrator.py",
            "sub_stage_orchestrator.py",
            "thin_slice_orchestrator.py",
            "__init__.py",
        ]

        for file in expected_files:
            assert os.path.exists(
                os.path.join(
                    output_dir, file
                )
            ), f"File {file} was not created"



from bclearer_core.common_knowledge.adjustment_operation_types import (
    AdjustmentOperationTypes,
)
from bclearer_core.configurations.adjustment_operation_configurations import (
    AdjustmentOperationConfigurations,
)
from bclearer_core.configurations.adjustment_operations_substage_configurations import (
    AdjustmentOperationsSubstageConfigurations,
)
from bclearer_core.configurations.attribute_to_association_adjustment_operation_configurations import (
    AttributeToAssociationAdjustmentOperationConfigurations,
)
from bclearer_core.substages.operations.a_load.content_operations.runners.hdf5_to_content_universe_loader import (
    load_hdf5_model_to_content_universe,
)
from bclearer_core.substages.operations.b_evolve.adjustment_operations.convert_attributes_to_associations.attributes_to_associations_converter import (
    convert_attributes_to_associations,
)
from bclearer_core.substages.operations.b_evolve.adjustment_operations.remove_attributes.attribute_remover import (
    remove_attributes,
)
from bclearer_core.substages.operations.b_evolve.common.universes_merge_registers import (
    UniversesMergeRegisters,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.nf_ea_com_universes import (
    NfEaComUniverses,
)
from bclearer_interop_services.ea_interop_service.session.orchestrators.ea_tools_session_managers import (
    EaToolsSessionManagers,
)
from bclearer_interop_services.ea_interop_service.session.processes.creators.empty_nf_ea_com_universe_creator import (
    create_empty_nf_ea_universe,
)


class AdjustmentOperationsSubstages:
    def __init__(
        self,
        ea_tools_session_manager: EaToolsSessionManagers,
        adjustment_operations_substage_configuration: AdjustmentOperationsSubstageConfigurations,
        content_universe: NfEaComUniverses,
    ):
        self.ea_tools_session_manager = (
            ea_tools_session_manager
        )

        self.adjustment_operations_substage_configuration = adjustment_operations_substage_configuration

        self.content_universe = (
            content_universe
        )

    def __enter__(self):
        return self

    def __exit__(
        self,
        exception_type,
        exception_value,
        traceback,
    ):
        pass

    def run(self) -> NfEaComUniverses:
        current_content_universe = (
            self.content_universe
        )

        for adjustment_operation_configuration in (
            self.adjustment_operations_substage_configuration.operation_configurations
        ):
            current_output_universe = self.__run_operation(
                content_universe=current_content_universe,
                adjustment_operation_configuration=adjustment_operation_configuration,
            )

            current_content_universe = (
                current_output_universe
            )

        return current_content_universe

    def __run_operation(
        self,
        content_universe: NfEaComUniverses,
        adjustment_operation_configuration: AdjustmentOperationConfigurations,
    ) -> NfEaComUniverses:
        adjustment_universe = load_hdf5_model_to_content_universe(
            ea_tools_session_manager=self.ea_tools_session_manager,
            load_hdf5_model_configuration=adjustment_operation_configuration.adjustment_universe_load_hdf5_model_configuration,
        )

        universes_merge_register = UniversesMergeRegisters(
            universe_1=content_universe,
            universe_2=adjustment_universe,
            context=self.__class__.__name__,
        )

        output_universe = create_empty_nf_ea_universe(
            ea_tools_session_manager=self.ea_tools_session_manager,
            short_name=adjustment_operation_configuration.output_universe_short_name,
        )

        if (
            adjustment_operation_configuration.operation_type
            == AdjustmentOperationTypes.REMOVE_ATTRIBUTES
        ):
            remove_attributes(
                content_universe=universes_merge_register.primary_universe,
                adjustment_universe=universes_merge_register.aligned_universe,
                output_universe=output_universe,
            )

        elif (
            adjustment_operation_configuration.operation_type
            == AdjustmentOperationTypes.CONVERT_ATTRIBUTES_TO_ASSOCIATIONS
        ):
            if not isinstance(
                adjustment_operation_configuration,
                AttributeToAssociationAdjustmentOperationConfigurations,
            ):
                raise TypeError(
                    "adjustment_operation_configuration is not AttributeToAssociationAdjustmentOperationConfigurations",
                )

            convert_attributes_to_associations(
                content_universe=universes_merge_register.primary_universe,
                adjustment_universe=universes_merge_register.aligned_universe,
                output_universe=output_universe,
                direction=adjustment_operation_configuration.direction,
                package_name=adjustment_operation_configuration.package_name,
                # attribute_to_association_operation_subtype=adjustment_operation_configuration.attribute_to_association_operation_subtype,
            )

        return output_universe



from bclearer_core.nf.types.nf_column_types import (
    NfColumnTypes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.column_types.nf_ea_com_column_types import (
    NfEaComColumnTypes,
)
from pandas import DataFrame


# TODO: This method may no longer be needed once the caller uses table ea_full_packages
def get_list_of_all_contained_packages_of_package(
    ea_packages_table: DataFrame,
    package_nf_uuid: str,
    set_of_all_packages_and_subpackages: set,
):
    list_of_contained_packages = ea_packages_table[
        ea_packages_table[
            NfColumnTypes.NF_UUIDS.column_name
        ]
        == package_nf_uuid
    ][
        NfEaComColumnTypes.PACKAGES_CONTAINED_EA_PACKAGES.column_name
    ].tolist()[
        0
    ]

    set_of_contained_packages = set(
        list_of_contained_packages,
    )

    if (
        len(set_of_contained_packages)
        == 0
    ):
        return set_of_all_packages_and_subpackages

    set_of_all_packages_and_subpackages = set_of_all_packages_and_subpackages.union(
        set_of_contained_packages,
    )

    for (
        subpackage
    ) in set_of_contained_packages:
        set_of_subpackages = get_list_of_all_contained_packages_of_package(
            ea_packages_table=ea_packages_table,
            package_nf_uuid=subpackage,
            set_of_all_packages_and_subpackages=set_of_all_packages_and_subpackages,
        )

        set_of_all_packages_and_subpackages = set_of_all_packages_and_subpackages.union(
            set_of_subpackages,
        )

    return set_of_all_packages_and_subpackages



"""Command-line interface for the bclearer pipeline builder.

This CLI tool helps create and update bclearer pipeline structures based on configuration.
The pipeline structure is generated entirely from scratch - no external template is needed.

Usage examples:
  # Generate a sample configuration file
  pipeline_builder sample --output my_config.json

  # Create a new pipeline from configuration file
  pipeline_builder create --config my_config.json

  # Create a pipeline interactively
  pipeline_builder create --interactive

  # Create a pipeline in a specific output directory
  pipeline_builder create --config my_config.json --output /path/to/output/directory

  # Update an existing pipeline with new components
  pipeline_builder update --config updated_config.json --pipeline path/to/domain_name_pipelines

  # Show detailed help
  pipeline_builder help
"""

import argparse
import json
import os
import sys
from pathlib import Path
from typing import Dict, Optional

from bclearer_core.pipeline_builder.generator import (
    generate_pipeline,
    update_pipeline,
)
from bclearer_core.pipeline_builder.schema import (
    get_sample_config,
)
from bclearer_core.pipeline_builder.template_extractor import (
    update_templates_from_pipeline,
)


def create_parser() -> (
    argparse.ArgumentParser
):
    """
    Create argument parser for the CLI.

    Returns:
        ArgumentParser object
    """
    parser = argparse.ArgumentParser(
        description="BClearer Pipeline Builder CLI"
    )
    subparsers = parser.add_subparsers(
        dest="command",
        help="Command to execute",
    )

    # help command
    help_parser = subparsers.add_parser(
        "help",
        help="Show detailed help and configuration guide",
    )

    # update-templates command
    update_templates_parser = subparsers.add_parser(
        "update-templates",
        help="Update template constants from a pipeline template",
    )
    update_templates_parser.add_argument(
        "--template-path",
        "-t",
        required=True,
        help="Path to the template pipeline directory",
    )
    update_templates_parser.add_argument(
        "--output",
        "-o",
        required=False,
        default=None,
        help="Output directory for templates (default: bclearer_core/pipeline_builder/templates)",
    )

    # create command
    create_parser = subparsers.add_parser(
        "create",
        help="Create a new pipeline",
    )
    create_parser.add_argument(
        "--config",
        "-c",
        required=False,
        help="Path to the pipeline configuration JSON file",
    )
    # Template parameter removed - structure is generated from scratch
    create_parser.add_argument(
        "--output",
        "-o",
        required=False,
        default=os.getcwd(),
        help="Base path where the new pipeline will be created (default: current directory)",
    )
    create_parser.add_argument(
        "--interactive",
        "-i",
        action="store_true",
        help="Run in interactive mode to create pipeline configuration",
    )

    # update command
    update_parser = subparsers.add_parser(
        "update",
        help="Update an existing pipeline",
    )
    update_parser.add_argument(
        "--config",
        "-c",
        required=True,
        help="Path to the pipeline configuration JSON file",
    )
    update_parser.add_argument(
        "--pipeline",
        "-p",
        required=True,
        help="Path to the existing pipeline directory",
    )

    # sample command
    sample_parser = subparsers.add_parser(
        "sample",
        help="Generate a sample configuration file",
    )
    sample_parser.add_argument(
        "--output",
        "-o",
        required=False,
        default="pipeline_config_sample.json",
        help="Output file path for sample configuration (default: pipeline_config_sample.json)",
    )

    return parser


def interactive_config() -> Dict:
    """
    Create pipeline configuration interactively.

    Returns:
        Dictionary with pipeline configuration
    """
    print(
        "=== BClearer Pipeline Builder Interactive Configuration ==="
    )

    domain_name = input("Domain name: ")

    pipelines = []
    pipeline_count = int(
        input("Number of pipelines: ")
    )

    for i in range(pipeline_count):
        pipeline_name = input(
            f"Pipeline {i+1} name: "
        )

        thin_slices = []
        slice_count = int(
            input(
                f"Number of thin slices for pipeline '{pipeline_name}': "
            )
        )

        for j in range(slice_count):
            slice_name = input(
                f"Thin slice {j+1} name for pipeline '{pipeline_name}': "
            )

            # Create standard stages
            stages = []
            for stage_info in [
                (
                    "1c_collect",
                    "Collect",
                ),
                ("2l_load", "Load"),
                ("3e_evolve", "Evolve"),
                (
                    "4a_assimilate",
                    "Assimilate",
                ),
                ("5r_reuse", "Reuse"),
            ]:
                (
                    stage_name,
                    stage_desc,
                ) = stage_info

                print(
                    f"\nStage: {stage_desc} ({stage_name})"
                )
                include_stage = (
                    input(
                        f"Include {stage_desc} stage? (y/n): "
                    ).lower()
                    == "y"
                )

                if include_stage:
                    b_units = []
                    has_b_units = (
                        input(
                            f"Does {stage_desc} stage have direct b_units? (y/n): "
                        ).lower()
                        == "y"
                    )

                    if has_b_units:
                        b_units_input = input(
                            "Enter b_unit names (comma-separated, e.g., ca,cb): "
                        )
                        b_units = [
                            unit.strip()
                            for unit in b_units_input.split(
                                ","
                            )
                            if unit.strip()
                        ]

                    sub_stages = []
                    has_sub_stages = (
                        input(
                            f"Does {stage_desc} stage have sub-stages? (y/n): "
                        ).lower()
                        == "y"
                    )

                    if has_sub_stages:
                        sub_stage_count = int(
                            input(
                                f"Number of sub-stages for {stage_desc}: "
                            )
                        )

                        for k in range(
                            sub_stage_count
                        ):
                            sub_stage_name = input(
                                f"Sub-stage {k+1} name for {stage_desc}: "
                            )

                            sub_b_units = (
                                []
                            )
                            sub_b_units_input = input(
                                f"Enter b_unit names for sub-stage {sub_stage_name} (comma-separated): "
                            )
                            sub_b_units = [
                                unit.strip()
                                for unit in sub_b_units_input.split(
                                    ","
                                )
                                if unit.strip()
                            ]

                            sub_stages.append(
                                {
                                    "name": sub_stage_name,
                                    "b_units": sub_b_units,
                                }
                            )

                    stages.append(
                        {
                            "name": stage_name,
                            "sub_stages": sub_stages,
                            "b_units": b_units,
                        }
                    )

            thin_slices.append(
                {
                    "name": slice_name,
                    "stages": stages,
                }
            )

        pipelines.append(
            {
                "name": pipeline_name,
                "thin_slices": thin_slices,
            }
        )

    return {
        "domain_name": domain_name,
        "pipelines": pipelines,
    }


def save_sample_config(
    output_path: str,
) -> None:
    """
    Save a sample configuration file.

    Args:
        output_path: Path to save the sample configuration file
    """
    sample_config = get_sample_config()

    with open(output_path, "w") as f:
        json.dump(
            sample_config, f, indent=2
        )

    print(
        f"Sample configuration saved to '{output_path}'"
    )


def show_detailed_help() -> None:
    """
    Show detailed help about pipeline configuration structure and usage.
    """
    help_text = """
=== BClearer Pipeline Builder - Detailed Help ===

COMMANDS:
  create    - Create a new pipeline structure from configuration (no template needed)
  update    - Update an existing pipeline with new components
  sample    - Generate a sample configuration file
  help      - Show this detailed help message

CONFIGURATION STRUCTURE:
  The pipeline configuration is a JSON file with the following structure:

  {
    "domain_name": "example_domain",
    "pipelines": [
      {
        "name": "example_pipeline",
        "thin_slices": [
          {
            "name": "example_thin_slice",
            "stages": [
              {
                "name": "1c_collect",
                "sub_stages": [
                  {
                    "name": "sub_stage_1",
                    "b_units": ["example_b_unit"]
                  }
                ],
                "b_units": ["collector_b_unit"]
              },
              // More stages: 2l_load, 3e_evolve, 4a_assimilate, 5r_reuse
            ]
          }
        ]
      }
    ]
  }

COMPONENT DETAILS:
  domain_name - The domain name for which pipelines are being built
  pipelines   - List of pipelines within the domain
  thin_slices - List of thin slices within a pipeline
  stages      - List of stages within a thin slice (standard stages: 1c_collect, 2l_load, 3e_evolve, 4a_assimilate, 5r_reuse)
  sub_stages  - Optional list of sub-stages within a stage
  b_units     - List of b_units (basic units of work) within a stage or sub-stage

CREATING A NEW PIPELINE:
  1. Generate a sample configuration:
     pipeline_builder sample --output my_config.json

  2. Edit the configuration file to match your requirements

  3. Create the pipeline structure:
     pipeline_builder create --config my_config.json

  4. Specify an output directory (optional):
     pipeline_builder create --config my_config.json --output /path/to/output/directory

  Alternatively, use interactive mode:
     pipeline_builder create --interactive

UPDATING AN EXISTING PIPELINE:
  1. Modify your configuration to add new components
     - Add new pipelines, thin slices, stages, sub-stages, or b_units
     - The domain_name must match the existing pipeline
     - Existing components will not be modified

  2. Update the pipeline structure:
     pipeline_builder update --config updated_config.json --pipeline path/to/domain_name_pipelines

TECHNICAL DETAILS:
  - The pipeline structure is generated completely from scratch - no external template is required
  - The generator creates all necessary directories, files, and code based on your configuration
  - This makes the tool portable and easy to use in any environment

TEMPLATE MANAGEMENT:
  - The pipeline generator uses template constants stored in the code
  - To update these templates from an existing pipeline template:
     pipeline_builder update-templates --template-path /path/to/template_pipeline

  - This allows you to maintain your code generation templates separately
  - Workflow:
    1. Make changes to the template pipeline code
    2. Run the update-templates command to extract the new templates
    3. The generator will now use the updated templates for all future pipelines

For more information, visit the BClearer documentation.
"""
    print(help_text)


def run_cli() -> None:
    """Run the bclearer pipeline builder CLI."""
    parser = create_parser()
    args = parser.parse_args()

    if not args.command:
        parser.print_help()
        return

    if args.command == "help":
        show_detailed_help()
        return

    if args.command == "sample":
        save_sample_config(args.output)
        return

    if args.command == "create":
        config = None

        if args.interactive:
            config = (
                interactive_config()
            )
        elif args.config:
            try:
                with open(
                    args.config, "r"
                ) as f:
                    config = json.load(
                        f
                    )
            except Exception as e:
                print(
                    f"Error loading configuration file: {str(e)}"
                )
                return
        else:
            print(
                "Error: Either --config or --interactive must be specified"
            )
            return

        try:
            pipeline_path = (
                generate_pipeline(
                    config,
                    args.output,
                )
            )
            print(
                f"Pipeline created successfully at '{pipeline_path}'"
            )
        except Exception as e:
            print(
                f"Error creating pipeline: {str(e)}"
            )
            return

    elif args.command == "update":
        try:
            with open(
                args.config, "r"
            ) as f:
                config = json.load(f)
        except Exception as e:
            print(
                f"Error loading configuration file: {str(e)}"
            )
            return

        try:
            pipeline_path = (
                update_pipeline(
                    config,
                    args.pipeline,
                )
            )
            print(
                f"Pipeline updated successfully at '{pipeline_path}'"
            )
        except Exception as e:
            print(
                f"Error updating pipeline: {str(e)}"
            )
            return

    elif (
        args.command
        == "update-templates"
    ):
        template_path = (
            args.template_path
        )

        if not os.path.exists(
            template_path
        ):
            print(
                f"Template path not found: {template_path}"
            )
            return

        if not os.path.isdir(
            template_path
        ):
            print(
                f"Template path is not a directory: {template_path}"
            )
            return

        # If output directory is not specified, use the default
        output_dir = args.output
        if output_dir is None:
            # Get the directory where this script is located
            script_dir = (
                os.path.dirname(
                    os.path.abspath(
                        __file__
                    )
                )
            )
            output_dir = os.path.join(
                script_dir, "templates"
            )

        try:
            update_templates_from_pipeline(
                template_path,
                output_dir,
            )
            print(
                f"Templates updated successfully in '{output_dir}'"
            )
        except Exception as e:
            print(
                f"Error updating templates: {str(e)}"
            )
            return


if __name__ == "__main__":
    run_cli()



import pandas
from bclearer_core.substages.operations.b_evolve.adjustment_operations.convert_attributes_to_associations.attribute_converter import (
    convert_attribute,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.collection_types.nf_ea_com_collection_types import (
    NfEaComCollectionTypes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.nf_ea_com_universes import (
    NfEaComUniverses,
)
from bclearer_interop_services.ea_interop_service.nf_ea_common.common_knowledge.ea_association_direction_types import (
    EaAssociationDirectionTypes,
)
from pandas import DataFrame


def convert_attributes(
    output_universe: NfEaComUniverses,
    attributes_to_convert: DataFrame,
    direction: EaAssociationDirectionTypes,
    package_nf_uuid: str,
) -> None:
    (
        new_classifiers_dictionary,
        new_connectors_dictionary,
    ) = __get_new_objects(
        attributes_to_convert=attributes_to_convert,
        direction=direction,
        package_nf_uuid=package_nf_uuid,
    )

    output_collections_dictionary = (
        output_universe.nf_ea_com_registry.dictionary_of_collections
    )

    __update_nf_ea_com_dictionary(
        ea_object_map=new_classifiers_dictionary,
        nf_ea_com_dictionary=output_collections_dictionary,
        nf_ea_collection_type=NfEaComCollectionTypes.EA_CLASSIFIERS,
    )

    __update_nf_ea_com_dictionary(
        ea_object_map=new_connectors_dictionary,
        nf_ea_com_dictionary=output_collections_dictionary,
        nf_ea_collection_type=NfEaComCollectionTypes.EA_CONNECTORS,
    )


def __get_new_objects(
    attributes_to_convert: DataFrame,
    direction: EaAssociationDirectionTypes,
    package_nf_uuid: str,
) -> tuple:
    new_classifiers_dictionary = {}

    new_connectors_dictionary = {}

    for (
        attribute_to_convert_tuple
    ) in (
        attributes_to_convert.itertuples()
    ):
        (
            new_classifiers_dictionary,
            new_connectors_dictionary,
        ) = convert_attribute(
            attribute_to_convert_tuple=attribute_to_convert_tuple,
            new_classifiers_dictionary=new_classifiers_dictionary,
            new_connectors_dictionary=new_connectors_dictionary,
            direction=direction,
            package_nf_uuid=package_nf_uuid,
        )

    return (
        new_classifiers_dictionary,
        new_connectors_dictionary,
    )


def __update_nf_ea_com_dictionary(
    nf_ea_com_dictionary: dict,
    ea_object_map: dict,
    nf_ea_collection_type: NfEaComCollectionTypes,
):
    nf_ea_collection = (
        nf_ea_com_dictionary[
            nf_ea_collection_type
        ]
    )

    converted_nf_ea_collection = (
        pandas.DataFrame.from_dict(
            data=ea_object_map,
            orient="index",
        )
    )

    all_ea_objects = pandas.concat(
        objs=[
            nf_ea_collection,
            converted_nf_ea_collection,
        ],
        ignore_index=True,
        verify_integrity=False,
    )

    nf_ea_com_dictionary[
        nf_ea_collection_type
    ] = all_ea_objects


# TODO: review this code that breaks boson pipeline.
#
# import pandas
# from bclearer_core.common_knowledge.attribute_to_associations_operation_subtypes import (
#     AttributeToAssociationOperationSubtypes,
# )
# from bclearer_core.substages.operations.b_evolve.adjustment_operations.convert_attributes_to_associations.attribute_to_direct_foreign_table_association_converter import (
#     convert_attribute_to_direct_foreign_table_association,
# )
# from bclearer_core.substages.operations.b_evolve.adjustment_operations.convert_attributes_to_associations.attribute_to_subtype_foreign_table_association_converter import (
#     convert_attribute_to_subtype_foreign_table_association,
# )
# from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.collection_types.nf_ea_com_collection_types import (
#     NfEaComCollectionTypes,
# )
# from bclearer_interop_services.ea_interop_service.general.nf_ea.com.nf_ea_com_universes import (
#     NfEaComUniverses,
# )
# from bclearer_interop_services.ea_interop_service.nf_ea_common.common_knowledge.ea_association_direction_types import (
#     EaAssociationDirectionTypes,
# )
# from pandas import DataFrame
#
#
# def convert_attributes(
#     output_universe: NfEaComUniverses,
#     attributes_to_convert: DataFrame,
#     direction: EaAssociationDirectionTypes,
#     package_nf_uuid: str,
#     #attribute_to_association_operation_subtype: AttributeToAssociationOperationSubtypes,
# ) -> None:
#     (
#         new_classifiers_dictionary,
#         new_connectors_dictionary,
#     ) = __get_new_objects(
#         attributes_to_convert=attributes_to_convert,
#         direction=direction,
#         package_nf_uuid=package_nf_uuid,
#         #attribute_to_association_operation_subtype=attribute_to_association_operation_subtype,
#     )
#
#     output_collections_dictionary = (
#         output_universe.nf_ea_com_registry.dictionary_of_collections
#     )
#
#     __update_nf_ea_com_dictionary(
#         ea_object_map=new_classifiers_dictionary,
#         nf_ea_com_dictionary=output_collections_dictionary,
#         nf_ea_collection_type=NfEaComCollectionTypes.EA_CLASSIFIERS,
#     )
#
#     __update_nf_ea_com_dictionary(
#         ea_object_map=new_connectors_dictionary,
#         nf_ea_com_dictionary=output_collections_dictionary,
#         nf_ea_collection_type=NfEaComCollectionTypes.EA_CONNECTORS,
#     )
#
#
# def __get_new_objects(
#         attributes_to_convert: DataFrame,
#         direction: EaAssociationDirectionTypes,
#         package_nf_uuid: str) \
#         -> tuple:
#     new_classifiers_dictionary = \
#         {}
#
#     new_connectors_dictionary = \
#         {}
#
#     for attribute_to_convert_tuple in attributes_to_convert.itertuples():
#         new_classifiers_dictionary, new_connectors_dictionary = \
#             convert_attribute(
#                 attribute_to_convert_tuple=attribute_to_convert_tuple,
#                 new_classifiers_dictionary=new_classifiers_dictionary,
#                 new_connectors_dictionary=new_connectors_dictionary,
#                 direction=direction,
#                 package_nf_uuid=package_nf_uuid)
#
#     return \
#         new_classifiers_dictionary, new_connectors_dictionary
#
#
# def __update_nf_ea_com_dictionary(
#         nf_ea_com_dictionary: dict,
#         ea_object_map: dict,
#         nf_ea_collection_type: NfEaComCollectionTypes):
#     nf_ea_collection = \
#         nf_ea_com_dictionary[nf_ea_collection_type]
#
#     converted_nf_ea_collection = \
#         pandas.DataFrame.from_dict(
#             data=ea_object_map,
#             orient='index')
#
#     all_ea_objects = \
#         pandas.concat(
#             objs=[nf_ea_collection, converted_nf_ea_collection],
#             ignore_index=True,
#             verify_integrity=False)
#
#     nf_ea_com_dictionary[nf_ea_collection_type] = \
#         all_ea_objects
#
#
# #
# # def __get_new_objects(
# #     attributes_to_convert: DataFrame,
# #     direction: EaAssociationDirectionTypes,
# #     package_nf_uuid: str,
# #     #attribute_to_association_operation_subtype: AttributeToAssociationOperationSubtypes,
# # ) -> tuple:
# #     new_classifiers_dictionary = {}
# #
# #     new_connectors_dictionary = {}
# #
# #     # TODO: Add the cardinality to the connector's data
# #     for (
# #         attribute_to_convert_tuple
# #     ) in (
# #         attributes_to_convert.itertuples()
# #     ):
# #         (
# #             new_classifiers_dictionary,
# #             new_connectors_dictionary,
# #         ) = __convert_attribute_to_association_by_subtype(
# #             attribute_to_convert_tuple=attribute_to_convert_tuple,
# #             new_classifiers_dictionary=new_classifiers_dictionary,
# #             new_connectors_dictionary=new_connectors_dictionary,
# #             direction=direction,
# #             package_nf_uuid=package_nf_uuid,
# #             #attribute_to_association_operation_subtype=attribute_to_association_operation_subtype,
# #         )
# #
# #     return (
# #         new_classifiers_dictionary,
# #         new_connectors_dictionary,
# #     )
# #
# #
# # def __convert_attribute_to_association_by_subtype(
# #     attribute_to_convert_tuple: tuple,
# #     direction: EaAssociationDirectionTypes,
# #     new_classifiers_dictionary: dict,
# #     new_connectors_dictionary: dict,
# #     package_nf_uuid: str,
# #     #attribute_to_association_operation_subtype: AttributeToAssociationOperationSubtypes,
# # ) -> tuple:
# #     if (
# #         attribute_to_association_operation_subtype
# #         is AttributeToAssociationOperationSubtypes.SUBTYPE_OF_FOREIGN_TABLE
# #     ):
# #         (
# #             new_classifiers_dictionary,
# #             new_connectors_dictionary,
# #         ) = convert_attribute_to_subtype_foreign_table_association(
# #             attribute_to_convert_tuple=attribute_to_convert_tuple,
# #             new_classifiers_dictionary=new_classifiers_dictionary,
# #             new_connectors_dictionary=new_connectors_dictionary,
# #             direction=direction,
# #             package_nf_uuid=package_nf_uuid,
# #         )
# #
# #     elif (
# #         attribute_to_association_operation_subtype
# #         is AttributeToAssociationOperationSubtypes.DIRECT_FOREIGN_TABLE
# #     ):
# #         (
# #             new_classifiers_dictionary,
# #             new_connectors_dictionary,
# #         ) = convert_attribute_to_direct_foreign_table_association(
# #             attribute_to_convert_tuple=attribute_to_convert_tuple,
# #             new_classifiers_dictionary=new_classifiers_dictionary,
# #             new_connectors_dictionary=new_connectors_dictionary,
# #             direction=direction,
# #         )
# #
# #     return (
# #         new_classifiers_dictionary,
# #         new_connectors_dictionary,
# #     )
# #
# #
# # def __update_nf_ea_com_dictionary(
# #     nf_ea_com_dictionary: dict,
# #     ea_object_map: dict,
# #     nf_ea_collection_type: NfEaComCollectionTypes,
# # ):
# #     nf_ea_collection = (
# #         nf_ea_com_dictionary[
# #             nf_ea_collection_type
# #         ]
# #     )
# #
# #     converted_nf_ea_collection = (
# #         pandas.DataFrame.from_dict(
# #             data=ea_object_map,
# #             orient="index",
# #         )
# #     )
# #
# #     all_ea_objects = pandas.concat(
# #         objs=[
# #             nf_ea_collection,
# #             converted_nf_ea_collection,
# #         ],
# #         ignore_index=True,
# #         verify_integrity=False,
# #     )
# #
# #     nf_ea_com_dictionary[
# #         nf_ea_collection_type
# #     ] = all_ea_objects



from bclearer_core.common_knowledge.bclearer_additional_column_types import (
    BclearerAdditionalColumnTypes,
)
from bclearer_core.common_knowledge.bclearer_constants import (
    NAME_EXEMPLAR_ATTRIBUTE_NAME,
)
from bclearer_core.common_knowledge.digitialisation_level_stereotype_matched_ea_objects import (
    DigitalisationLevelStereotypeMatchedEaObjects,
)
from bclearer_core.constants.standard_constants import (
    DEFAULT_FOREIGN_TABLE_SUFFIX,
    DEFAULT_MASTER_TABLE_SUFFIX,
)
from bclearer_core.substages.operations.b_evolve.common.new_root_package_creator import (
    create_root_package,
)
from bclearer_core.substages.operations.b_evolve.convention_shift_operations.convention_shifters.separate_instances_and_exemplars.instances_and_exemplars_separator import (
    separate_instances_and_exemplars,
)
from bclearer_core.substages.operations.common.new_ea_objects_dictionary_creator import (
    create_new_ea_objects_dictionary,
)
from bclearer_core.substages.operations.common.nf_ea_com_universe_updater import (
    update_nf_ea_com_universe_with_dictionary,
)
from bclearer_core.substages.operations.common.nf_uuid_from_ea_guid_from_collection_getter import (
    get_nf_uuid_from_ea_guid_from_collection,
)
from bclearer_interop_services.dataframe_service.dataframe_helpers.dataframe_mergers import (
    left_merge_dataframes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.collection_types.nf_ea_com_collection_types import (
    NfEaComCollectionTypes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.column_types.nf_ea_com_column_types import (
    NfEaComColumnTypes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.nf_ea_com_universes import (
    NfEaComUniverses,
)
from bclearer_interop_services.ea_interop_service.nf_ea_common.common_knowledge.ea_connector_types import (
    EaConnectorTypes,
)
from bclearer_orchestration_services.reporting_service.reporters.log_with_datetime import (
    log_message,
)


def shift_convention_separate_standard_instances_and_exemplars(
    content_universe: NfEaComUniverses,
    output_universe: NfEaComUniverses,
    package_name: str,
) -> NfEaComUniverses:
    log_message(
        message="Separate Standard Instances and Exemplars - started",
    )

    output_universe.nf_ea_com_registry.dictionary_of_collections = (
        content_universe.nf_ea_com_registry.dictionary_of_collections.copy()
    )

    __separate_instances_and_exemplars(
        nf_ea_com_universe=output_universe,
        package_name=package_name,
    )

    log_message(
        message="Separate Standard Instances and Exemplars - finished",
    )

    return output_universe


def __separate_instances_and_exemplars(
    nf_ea_com_universe: NfEaComUniverses,
    package_name: str,
):
    package_nf_uuid = create_root_package(
        nf_ea_com_universe=nf_ea_com_universe,
        package_name=package_name,
    )

    new_ea_objects_dictionary = (
        create_new_ea_objects_dictionary()
    )

    attributes_grouped_by_name_instance_type_dictionary = __get_attributes_grouped_by_name_instance_type_dictionary(
        nf_ea_com_universe=nf_ea_com_universe,
    )

    digitalisation_level_stereotype_nf_uuid = get_nf_uuid_from_ea_guid_from_collection(
        nf_ea_com_universe=nf_ea_com_universe,
        collection_type=NfEaComCollectionTypes.EA_STEREOTYPES,
        ea_guid=DigitalisationLevelStereotypeMatchedEaObjects.DIGITALISATION_LEVEL_1_CLASS_STEREOTYPE.ea_guid,
    )

    for (
        name_instance_type_nf_uuid,
        ea_attributes,
    ) in (
        attributes_grouped_by_name_instance_type_dictionary.items()
    ):
        separate_instances_and_exemplars(
            nf_ea_com_universe=nf_ea_com_universe,
            name_instance_type_nf_uuid=name_instance_type_nf_uuid,
            ea_attributes=ea_attributes,
            new_ea_objects_dictionary=new_ea_objects_dictionary,
            package_nf_uuid=package_nf_uuid,
            digitalisation_level_stereotype_nf_uuid=digitalisation_level_stereotype_nf_uuid,
        )

    update_nf_ea_com_universe_with_dictionary(
        nf_ea_com_universe=nf_ea_com_universe,
        new_ea_objects_dictionary=new_ea_objects_dictionary,
    )


def __get_attributes_grouped_by_name_instance_type_dictionary(
    nf_ea_com_universe: NfEaComUniverses,
) -> dict:
    ea_attributes = (
        nf_ea_com_universe.nf_ea_com_registry.get_ea_attributes()
    )

    standard_name_exemplar_attributes = ea_attributes[
        ea_attributes[
            NfEaComColumnTypes.EXPLICIT_OBJECTS_EA_OBJECT_NAME.column_name
        ]
        == NAME_EXEMPLAR_ATTRIBUTE_NAME
    ]

    ea_connectors = (
        nf_ea_com_universe.nf_ea_com_registry.get_ea_connectors()
    )

    ea_dependencies = ea_connectors[
        ea_connectors[
            NfEaComColumnTypes.CONNECTORS_ELEMENT_TYPE_NAME.column_name
        ]
        == EaConnectorTypes.DEPENDENCY.type_name
    ]

    extended_standard_name_exemplar_attributes = left_merge_dataframes(
        master_dataframe=standard_name_exemplar_attributes,
        master_dataframe_key_columns=[
            NfEaComColumnTypes.ELEMENT_COMPONENTS_CONTAINING_EA_CLASSIFIER.column_name,
        ],
        merge_suffixes=[
            DEFAULT_FOREIGN_TABLE_SUFFIX,
            DEFAULT_MASTER_TABLE_SUFFIX,
        ],
        foreign_key_dataframe=ea_dependencies,
        foreign_key_dataframe_fk_columns=[
            NfEaComColumnTypes.ELEMENTS_SUPPLIER_PLACE1_END_CONNECTORS.column_name,
        ],
        foreign_key_dataframe_other_column_rename_dictionary={
            NfEaComColumnTypes.ELEMENTS_CLIENT_PLACE2_END_CONNECTORS.column_name: BclearerAdditionalColumnTypes.NAME_INSTANCE_TYPE_NF_UUIDS.column_name,
        },
    )

    name_instance_type_nf_uuids = set(
        extended_standard_name_exemplar_attributes[
            BclearerAdditionalColumnTypes.NAME_INSTANCE_TYPE_NF_UUIDS.column_name
        ],
    )

    attributes_grouped_by_name_instance_type_dictionary = (
        dict()
    )

    for (
        name_instance_type_nf_uuid
    ) in name_instance_type_nf_uuids:
        name_instance_type_ea_attributes = extended_standard_name_exemplar_attributes[
            extended_standard_name_exemplar_attributes[
                BclearerAdditionalColumnTypes.NAME_INSTANCE_TYPE_NF_UUIDS.column_name
            ]
            == name_instance_type_nf_uuid
        ]

        attributes_grouped_by_name_instance_type_dictionary[
            name_instance_type_nf_uuid
        ] = name_instance_type_ea_attributes

    return attributes_grouped_by_name_instance_type_dictionary



from bclearer_core.common_knowledge.matched_objects import (
    MatchedEaObjects,
)


class BespokeInstanceToExemplarConfigurationObjects:
    def __init__(
        self,
        matched_name_instance_instance: MatchedEaObjects,
        name_exemplar_attribute_name: str,
    ):
        self.matched_name_instance_instance = matched_name_instance_instance

        self.name_exemplar_attribute_name = (
            name_exemplar_attribute_name
        )

    def __enter__(self):
        return self

    def __exit__(
        self,
        exception_type,
        exception_value,
        traceback,
    ):
        pass



from bclearer_core.nf.types.nf_column_types import (
    NfColumnTypes,
)
from bclearer_interop_services.dataframe_service.dataframe_helpers.dataframe_mergers import (
    inner_merge_dataframes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.collection_types.nf_ea_com_collection_types import (
    NfEaComCollectionTypes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.column_types.nf_ea_com_column_types import (
    NfEaComColumnTypes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.nf_ea_com_universes import (
    NfEaComUniverses,
)
from bclearer_interop_services.ea_interop_service.nf_ea_common.common_knowledge.ea_connector_types import (
    EaConnectorTypes,
)
from bclearer_interop_services.ea_interop_service.nf_ea_common.common_knowledge.ea_property_types import (
    EaPropertyTypes,
)


def add_stereotype_to_instances_associations(
    nf_ea_com_universe: NfEaComUniverses,
    type_nf_uuid: str,
    stereotype_nf_uuid: str,
) -> None:
    connectors = nf_ea_com_universe.nf_ea_com_registry.dictionary_of_collections[
        NfEaComCollectionTypes.EA_CONNECTORS
    ]

    connectors_to_type = connectors[
        connectors[
            NfEaComColumnTypes.ELEMENTS_CLIENT_PLACE2_END_CONNECTORS.column_name
        ]
        == type_nf_uuid
    ]

    dependencies_to_type = connectors_to_type[
        connectors_to_type[
            NfEaComColumnTypes.CONNECTORS_ELEMENT_TYPE_NAME.column_name
        ]
        == EaConnectorTypes.DEPENDENCY.type_name
    ]

    associations = connectors[
        connectors[
            NfEaComColumnTypes.CONNECTORS_ELEMENT_TYPE_NAME.column_name
        ]
        == EaConnectorTypes.ASSOCIATION.type_name
    ]

    linked_relations = inner_merge_dataframes(
        master_dataframe=dependencies_to_type,
        master_dataframe_key_columns=[
            NfEaComColumnTypes.ELEMENTS_SUPPLIER_PLACE1_END_CONNECTORS.column_name,
        ],
        merge_suffixes=[
            "_dependency",
            "_association",
        ],
        foreign_key_dataframe=associations,
        foreign_key_dataframe_fk_columns=[
            NfEaComColumnTypes.ELEMENTS_SUPPLIER_PLACE1_END_CONNECTORS.column_name,
        ],
        foreign_key_dataframe_other_column_rename_dictionary={
            NfColumnTypes.NF_UUIDS.column_name: NfEaComColumnTypes.STEREOTYPE_CLIENT_NF_UUIDS.column_name,
        },
    )

    new_stereotype_usages = linked_relations.filter(
        items=[
            NfEaComColumnTypes.STEREOTYPE_CLIENT_NF_UUIDS.column_name,
        ],
    )

    new_stereotype_usages[
        NfEaComColumnTypes.STEREOTYPE_PROPERTY_TYPE.column_name
    ] = (
        EaPropertyTypes.CONNECTOR_PROPERTY.type_name
    )

    new_stereotype_usages[
        "stereotype_nf_uuids"
    ] = stereotype_nf_uuid

    nf_ea_com_universe.nf_ea_com_registry.update(
        collection_type=NfEaComCollectionTypes.STEREOTYPE_USAGE,
        new_collection=new_stereotype_usages,
    )



import os.path

from bclearer_interop_services.file_system_service.objects.file_system_objects import (
    FileSystemObjects,
)
from bclearer_interop_services.file_system_service.objects.folders_latest import (
    Folders,
)
from bclearer_orchestration_services.datetime_service.time_helpers.time_getter import (
    now_time_as_string_for_files,
)


class BConfigurations:
    default_string_empty = str()

    default_data_inspection = True

    # INPUT_ROOT_FOLDER = \
    #     Folders(
    #         absolute_path_string=default_string_empty)

    INPUT_FILE_SYSTEM_OBJECT = FileSystemObjects(
        absolute_path_string=default_string_empty
    )

    OUTPUT_ROOT_FOLDER = Folders(
        absolute_path_string=default_string_empty
    )

    APP_RUN_OUTPUT_FOLDER_NAME = (
        default_string_empty
    )

    APP_RUN_OUTPUT_FOLDER = Folders(
        absolute_path_string=default_string_empty
    )

    APP_RUN_TIME_AS_STRING = (
        default_string_empty
    )

    APP_RUN_OUTPUT_FOLDER_PREFIX = (
        default_string_empty
    )

    APP_RUN_OUTPUT_FOLDER_SUFFIX = (
        default_string_empty
    )

    ENABLE_CSV_FILE_INSPECTION = (
        default_data_inspection
    )

    ENABLE_MS_ACCESS_DATABASE_INSPECTION = (
        default_data_inspection
    )

    ENABLE_SQLITE_DATABASE_INSPECTION = (
        default_data_inspection
    )

    @staticmethod
    def set_up_app_run_output_folder_name() -> (
        None
    ):
        datetime_stamp = (
            now_time_as_string_for_files()
        )

        output_folder_name = (
            datetime_stamp
        )

        if (
            BConfigurations.APP_RUN_OUTPUT_FOLDER_PREFIX
        ):
            output_folder_name = (
                BConfigurations.APP_RUN_OUTPUT_FOLDER_PREFIX
                + "_"
                + output_folder_name
            )

        if (
            BConfigurations.APP_RUN_OUTPUT_FOLDER_SUFFIX
        ):
            output_folder_name = (
                output_folder_name
                + "_"
                + BConfigurations.APP_RUN_OUTPUT_FOLDER_SUFFIX
            )

        BConfigurations.APP_RUN_OUTPUT_FOLDER_NAME = (
            output_folder_name
        )

        BConfigurations.APP_RUN_TIME_AS_STRING = (
            datetime_stamp
        )

    @staticmethod
    def set_app_run_output_folder() -> (
        None
    ):
        BConfigurations.APP_RUN_OUTPUT_FOLDER = Folders(
            absolute_path_string=os.path.join(
                BConfigurations.OUTPUT_ROOT_FOLDER.absolute_path_string,
                BConfigurations.APP_RUN_OUTPUT_FOLDER_NAME,
            )
        )

    @staticmethod
    def persist_dataframe_to_access_if_enabled() -> (
        None
    ):
        BConfigurations.ENABLE_MS_ACCESS_DATABASE_INSPECTION = (
            False
        )

        # TODO: Needs to work for NT
        # if os.name == 'nt':
        #     export_all_csv_files_from_folder_to_access(
        #             csv_folder=output_folder,
        #             database_already_exists=False,
        #             new_database_name_if_not_exists=output_folder.base_name + '_register_')



"""Template for pipelines runner."""

PIPELINES_RUNNER_TEMPLATE_HEADER = """from bclearer_orchestration_services.reporting_service.wrappers.run_and_log_function_wrapper_latest import (
    run_and_log_function,
)
"""

PIPELINE_IMPORT_TEMPLATE = """from bclearer_pipelines.{domain_name}.b_source.app_runners.runners.{pipeline_name}_runner import (
    run_{pipeline_name},
)
"""

PIPELINES_RUNNER_FUNCTION_START = """

@run_and_log_function()
def run_{domain_name}_b_clearer_pipelines() -> (
    None
):
"""

PIPELINE_RUNNER_CALL = """    run_{pipeline_name}()
"""

EMPTY_FUNCTION_BODY = """    pass
"""



from bclearer_core.common_knowledge.content_operation_types import (
    ContentOperationTypes,
)
from bclearer_core.common_knowledge.digitialisation_level_stereotype_matched_ea_objects import (
    DigitalisationLevelStereotypeMatchedEaObjects,
)
from bclearer_core.configurations.operation_configurations import (
    OperationConfigurations,
)


class ContentOperationConfigurations(
    OperationConfigurations,
):
    def __init__(
        self,
        content_operation_type: ContentOperationTypes,
        output_universe_short_name: str,
        default_digitalisation_level_stereotype: DigitalisationLevelStereotypeMatchedEaObjects = None,
    ):
        super().__init__(
            operation_type=content_operation_type,
        )

        self.output_universe_short_name = (
            output_universe_short_name
        )

        self.default_digitalisation_level_stereotype = default_digitalisation_level_stereotype

    def __enter__(self):
        return self

    def __exit__(
        self,
        exception_type,
        exception_value,
        traceback,
    ):
        pass



from bclearer_core.constants.standard_constants import (
    DEFAULT_FOREIGN_TABLE_SUFFIX,
    DEFAULT_MASTER_TABLE_SUFFIX,
    DEFAULT_NULL_VALUE,
)
from bclearer_core.nf.types.nf_column_types import (
    NfColumnTypes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.column_types.nf_ea_com_column_types import (
    NfEaComColumnTypes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.nf_ea_com_universes import (
    NfEaComUniverses,
)
from pandas import (
    DataFrame,
    concat,
    merge,
)


def create_aligned_to_primary_universe_nf_uuids_map(
    primary_universe: NfEaComUniverses,
    aligned_universe: NfEaComUniverses,
) -> dict:
    matched_by_ea_guid_nf_uuids_dataframe = __get_matched_by_ea_guid_mapping_dataframe(
        primary_universe=primary_universe,
        secondary_universe=aligned_universe,
    )

    secondary_to_primary_universe_nf_uuids_map = __get_secondary_to_primary_universe_nf_uuids_map(
        matched_by_ea_guid_nf_uuids_dataframe=matched_by_ea_guid_nf_uuids_dataframe,
    )

    return secondary_to_primary_universe_nf_uuids_map


def __get_matched_by_ea_guid_mapping_dataframe(
    primary_universe: NfEaComUniverses,
    secondary_universe: NfEaComUniverses,
) -> DataFrame:
    primary_universe_nf_uuids_to_ea_guids_table = __get_universe_nf_uuids_to_ea_guids_table(
        content_universe=primary_universe,
    )

    secondary_universe_nf_uuids_to_ea_guids_table = __get_universe_nf_uuids_to_ea_guids_table(
        content_universe=secondary_universe,
    )

    matched_by_ea_guid_mapping_dataframe = merge(
        left=secondary_universe_nf_uuids_to_ea_guids_table,
        right=primary_universe_nf_uuids_to_ea_guids_table,
        suffixes=(
            DEFAULT_MASTER_TABLE_SUFFIX,
            DEFAULT_FOREIGN_TABLE_SUFFIX,
        ),
        how="outer",
        on=NfEaComColumnTypes.EXPLICIT_OBJECTS_EA_GUID.column_name,
    )

    matched_by_ea_guid_mapping_dataframe.drop_duplicates(
        inplace=True,
    )

    return matched_by_ea_guid_mapping_dataframe


def __get_universe_nf_uuids_to_ea_guids_table(
    content_universe: NfEaComUniverses,
) -> DataFrame:
    universe_nf_uuids_to_ea_guids_table = DataFrame(
        columns=[
            NfColumnTypes.NF_UUIDS.column_name,
            NfEaComColumnTypes.EXPLICIT_OBJECTS_EA_GUID.column_name,
        ],
    )

    for (
        collection_type,
        collection,
    ) in (
        content_universe.nf_ea_com_registry.dictionary_of_collections.items()
    ):
        universe_nf_uuids_to_ea_guids_table = __add_nf_uuids_to_ea_guids_if_required(
            universe_nf_uuids_to_ea_guids_table=universe_nf_uuids_to_ea_guids_table,
            collection=collection,
        )

    return universe_nf_uuids_to_ea_guids_table


def __add_nf_uuids_to_ea_guids_if_required(
    universe_nf_uuids_to_ea_guids_table: DataFrame,
    collection: DataFrame,
) -> DataFrame:
    if (
        NfColumnTypes.NF_UUIDS.column_name
        not in collection.columns
    ):
        return universe_nf_uuids_to_ea_guids_table

    if (
        NfEaComColumnTypes.EXPLICIT_OBJECTS_EA_GUID.column_name
        not in collection.columns
    ):
        return universe_nf_uuids_to_ea_guids_table

    new_mappings = collection.filter(
        items=[
            NfColumnTypes.NF_UUIDS.column_name,
            NfEaComColumnTypes.EXPLICIT_OBJECTS_EA_GUID.column_name,
        ],
    )

    new_mappings = new_mappings.loc[
        new_mappings[
            NfEaComColumnTypes.EXPLICIT_OBJECTS_EA_GUID.column_name
        ]
        != DEFAULT_NULL_VALUE
    ]

    # Check if both DataFrames are non-empty before concatenating
    if (
        not new_mappings.empty
        and not universe_nf_uuids_to_ea_guids_table.empty
    ):
        universe_nf_uuids_to_ea_guids_table = concat(
            [
                universe_nf_uuids_to_ea_guids_table,
                new_mappings,
            ],
        )
    elif not new_mappings.empty:
        universe_nf_uuids_to_ea_guids_table = (
            new_mappings
        )

    universe_nf_uuids_to_ea_guids_table = (
        universe_nf_uuids_to_ea_guids_table.drop_duplicates()
    )

    universe_nf_uuids_to_ea_guids_table.reset_index(
        drop=True,
        inplace=True,
    )

    return universe_nf_uuids_to_ea_guids_table


# TODO: patched, needs to b removed

# def __add_nf_uuids_to_ea_guids_if_required(
#     universe_nf_uuids_to_ea_guids_table: DataFrame,
#     collection: DataFrame,
# ) -> DataFrame:
#     if (
#         NfColumnTypes.NF_UUIDS.column_name
#         not in collection.columns
#     ):
#         return universe_nf_uuids_to_ea_guids_table
#
#     if (
#         NfEaComColumnTypes.EXPLICIT_OBJECTS_EA_GUID.column_name
#         not in collection.columns
#     ):
#         return universe_nf_uuids_to_ea_guids_table
#
#     new_mappings = collection.filter(
#         items=[
#             NfColumnTypes.NF_UUIDS.column_name,
#             NfEaComColumnTypes.EXPLICIT_OBJECTS_EA_GUID.column_name,
#         ],
#     )
#
#     new_mappings = new_mappings.loc[
#         new_mappings[
#             NfEaComColumnTypes.EXPLICIT_OBJECTS_EA_GUID.column_name
#         ]
#         != DEFAULT_NULL_VALUE
#     ]
#
#     universe_nf_uuids_to_ea_guids_table = concat(
#         [
#             universe_nf_uuids_to_ea_guids_table,
#             new_mappings,
#         ],
#     )
#
#     universe_nf_uuids_to_ea_guids_table = (
#         universe_nf_uuids_to_ea_guids_table.drop_duplicates()
#     )
#
#     universe_nf_uuids_to_ea_guids_table.reset_index(
#         drop=True,
#         inplace=True,
#     )
#
#     return universe_nf_uuids_to_ea_guids_table


def __get_secondary_to_primary_universe_nf_uuids_map(
    matched_by_ea_guid_nf_uuids_dataframe: DataFrame,
) -> dict:
    matched_by_ea_guid_nf_uuids_dataframe_copy = (
        matched_by_ea_guid_nf_uuids_dataframe.copy()
    )

    key_column = (
        NfColumnTypes.NF_UUIDS.column_name
        + DEFAULT_MASTER_TABLE_SUFFIX
    )

    value_column = (
        NfColumnTypes.NF_UUIDS.column_name
        + DEFAULT_FOREIGN_TABLE_SUFFIX
    )

    matched_by_ea_guid_nf_uuids_dataframe_copy = matched_by_ea_guid_nf_uuids_dataframe_copy[
        [
            key_column,
            value_column,
        ]
    ]

    matched_by_ea_guid_nf_uuids_dataframe_copy.dropna(
        inplace=True,
    )

    secondary_to_primary_universe_nf_uuids_map = matched_by_ea_guid_nf_uuids_dataframe_copy.set_index(
        key_column,
    )[
        value_column
    ].to_dict()

    return secondary_to_primary_universe_nf_uuids_map



from enum import Enum


class PlaceNumberTypeCkIds(Enum):
    NotSet = 0

    PlaceNumberOnes = 1
    PlaceNumberTwos = 2
    PlaceNumberThrees = 3
    PlaceNumberFours = 4
    PlaceNumberFives = 5
    PlaceNumberSixes = 6



"""Template for b_unit creator and runner."""

B_UNIT_CREATOR_AND_RUNNER_TEMPLATE = """# TODO: Minimal version of the method developed in core graph mvp dev that should be promoted to a bCLEARer importable
#  library
def create_and_run_b_unit(
    b_unit_type,
) -> None:
    b_unit = b_unit_type()

    b_unit.run()
"""



from nf_common_base.b_source.common.infrastructure.nf.objects.registers.nf_verse_registers import (
    NfVerseRegisters,
)


class NfMultiverseRegisters(
    NfVerseRegisters
):
    def __init__(self):
        super().__init__()

    def __enter__(self):
        return self

    def __exit__(
        self,
        exception_type,
        exception_value,
        traceback,
    ):
        pass



from bclearer_core.substages.operations.common.classifier_adder import (
    add_new_classifier_to_dictionary,
)
from bclearer_core.substages.operations.common.connector_adder import (
    add_new_connector_to_dictionary,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.collection_types.nf_ea_com_collection_types import (
    NfEaComCollectionTypes,
)
from bclearer_interop_services.ea_interop_service.nf_ea_common.common_knowledge.ea_connector_types import (
    EaConnectorTypes,
)
from bclearer_interop_services.ea_interop_service.nf_ea_common.common_knowledge.ea_element_types import (
    EaElementTypes,
)


def add_new_connector_to_connector_to_dictionary(
    new_ea_objects_dictionary: dict,
    package_nf_uuid: str,
    connector_nf_uuid: str,
    class_nf_uuid: str,
) -> str:
    proxy_connector_nf_uuid = add_new_classifier_to_dictionary(
        new_classifier_dictionary=new_ea_objects_dictionary[
            NfEaComCollectionTypes.EA_CLASSIFIERS
        ],
        package_nf_uuid=package_nf_uuid,
        ea_element_type=EaElementTypes.PROXY_CONNECTOR,
        elements_classifier=connector_nf_uuid,
    )

    connector_nf_uuid = add_new_connector_to_dictionary(
        new_connector_dictionary=new_ea_objects_dictionary[
            NfEaComCollectionTypes.EA_CONNECTORS
        ],
        place_1_nf_uuid=proxy_connector_nf_uuid,
        place_2_nf_uuid=class_nf_uuid,
        connector_type=EaConnectorTypes.DEPENDENCY,
    )

    return connector_nf_uuid



"""Template for sub-stage orchestrator."""

SUB_STAGE_ORCHESTRATOR_HEADER = """from bclearer_orchestration_services.reporting_service.wrappers.run_and_log_function_wrapper_latest import (
    run_and_log_function,
)
"""

B_UNIT_CREATOR_IMPORT = """from bclearer_pipelines.{domain_name}.b_source.common.operations.b_units.b_unit_creator_and_runner import (
    create_and_run_b_unit,
)
"""

SUB_STAGE_B_UNIT_IMPORT_TEMPLATE = """from bclearer_pipelines.{domain_name}.b_source.{pipeline_name}.objects.b_units.{pipeline_name}_{stage_name}.{pipeline_name}_{stage_name}_{sub_stage_name}.{b_unit_lower}_b_units import (
    {b_unit_class}BUnits,
)
"""

SUB_STAGE_ORCHESTRATOR_FUNCTION_START = """

@run_and_log_function()
def orchestrate_{pipeline_name}_{stage_name}_{sub_stage_name}() -> (
    None
):
    __run_contained_bie_pipeline_components()


def __run_contained_bie_pipeline_components() -> (
    None
):
"""

CREATE_AND_RUN_B_UNIT_CALL = """    create_and_run_b_unit(
        b_unit_type={b_unit_class}BUnits
    )
"""

EMPTY_FUNCTION_BODY = """    pass
"""



from bclearer_core.substages.operations.b_evolve.common.universe_align_reporter import (
    report_collection_with_different_columns,
    report_universes_with_different_collection_types,
)
from bclearer_core.substages.operations.b_evolve.content_operations.merge_universes.nf_uuid_mapping_processes.merged_collection_nf_uuids_replacer import (
    replace_nf_uuids_in_collection,
)
from bclearer_core.substages.operations.b_evolve.content_operations.merge_universes.nf_uuid_mapping_processes.nf_uuid_mapping_creator import (
    create_aligned_to_primary_universe_nf_uuids_map,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.collection_types.nf_ea_com_collection_types import (
    NfEaComCollectionTypes,
)
from bclearer_orchestration_services.reporting_service.reporters.log_with_datetime import (
    log_message,
)
from pandas import DataFrame


def align_universes(
    universe_merge_register,
    context: str,
):
    primary_universe_name = (
        universe_merge_register.primary_universe.ea_repository.short_name
    )

    aligned_universe_name = (
        universe_merge_register.aligned_universe.ea_repository.short_name
    )

    log_message(
        message="In context "
        + context
        + " two universes are nf uuid aligned: "
        + primary_universe_name
        + " as primary and "
        + aligned_universe_name
        + " as aligned",
    )

    alignment_scope = __set_alignment_scope(
        universe_merge_register=universe_merge_register,
    )

    __align_universe_registries(
        alignment_scope=alignment_scope,
        universe_merge_register=universe_merge_register,
    )


def __set_alignment_scope(
    universe_merge_register,
) -> set:
    collection_types_in_primary_universe = set(
        universe_merge_register.primary_universe.nf_ea_com_registry.dictionary_of_collections.keys(),
    )

    collection_types_in_aligned_universe = set(
        universe_merge_register.aligned_universe.nf_ea_com_registry.dictionary_of_collections.keys(),
    )

    if (
        collection_types_in_primary_universe
        != collection_types_in_aligned_universe
    ):
        report_universes_with_different_collection_types(
            collection_types_in_primary_universe=collection_types_in_primary_universe,
            collection_types_in_aligned_universe=collection_types_in_aligned_universe,
        )

    return collection_types_in_primary_universe.intersection(
        collection_types_in_aligned_universe,
    )


def __align_universe_registries(
    universe_merge_register,
    alignment_scope: set,
):
    for (
        collection_type
    ) in alignment_scope:
        __check_column_consistency_in_alignment(
            collection_type=collection_type,
            universe_merge_register=universe_merge_register,
        )

    universe_merge_register.aligned_to_primary_universe_nf_uuids_map = create_aligned_to_primary_universe_nf_uuids_map(
        primary_universe=universe_merge_register.primary_universe,
        aligned_universe=universe_merge_register.aligned_universe,
    )

    for (
        collection_type,
        collection,
    ) in (
        universe_merge_register.aligned_universe.nf_ea_com_registry.dictionary_of_collections.items()
    ):
        __align_collection_in_universe(
            collection=collection,
            collection_type=collection_type,
            universe_merge_register=universe_merge_register,
        )


def __check_column_consistency_in_alignment(
    universe_merge_register,
    collection_type: NfEaComCollectionTypes,
):
    primary_universe_collection = universe_merge_register.primary_universe.nf_ea_com_registry.dictionary_of_collections[
        collection_type
    ]

    aligned_universe_collection = universe_merge_register.aligned_universe.nf_ea_com_registry.dictionary_of_collections[
        collection_type
    ]

    primary_universe_collection_columns = set(
        primary_universe_collection.columns,
    )

    aligned_universe_collection_columns = set(
        aligned_universe_collection.columns,
    )

    if (
        primary_universe_collection_columns
        != aligned_universe_collection_columns
    ):
        report_collection_with_different_columns(
            primary_universe_collection_columns=primary_universe_collection_columns,
            aligned_universe_collection_columns=aligned_universe_collection_columns,
            collection_type=collection_type,
        )


def __align_collection_in_universe(
    universe_merge_register,
    collection: DataFrame,
    collection_type: NfEaComCollectionTypes,
):
    updated_collection = (
        collection.copy()
    )

    for column in collection.columns:
        updated_collection = replace_nf_uuids_in_collection(
            column_name=column,
            nf_uuid_mapping_dictionary=universe_merge_register.aligned_to_primary_universe_nf_uuids_map,
            collection=updated_collection,
        )

    universe_merge_register.aligned_universe.nf_ea_com_registry.dictionary_of_collections[
        collection_type
    ] = updated_collection



from bclearer_core.nf.types.nf_column_types import (
    NfColumnTypes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.column_types.nf_ea_com_column_types import (
    NfEaComColumnTypes,
)
from bclearer_orchestration_services.identification_services.uuid_service.uuid_helpers.uuid_factory import (
    create_new_uuid,
)


def add_new_stereotype_to_dictionary(
    new_stereotype_dictionary: dict,
    stereotype_ea_guid: str,
    stereotype_name: str,
    stereotype_applies_to: str,
    stereotype_style: str,
) -> None:
    new_stereotype_uuid = (
        create_new_uuid()
    )

    stereotype_row_dictionary = {
        NfColumnTypes.NF_UUIDS.column_name: new_stereotype_uuid,
        NfEaComColumnTypes.EXPLICIT_OBJECTS_EA_GUID.column_name: stereotype_ea_guid,
        NfEaComColumnTypes.EXPLICIT_OBJECTS_EA_OBJECT_NAME.column_name: stereotype_name,
        NfEaComColumnTypes.STEREOTYPE_APPLIES_TOS.column_name: stereotype_applies_to,
        NfEaComColumnTypes.STEREOTYPE_STYLE.column_name: stereotype_style,
    }

    new_stereotype_dictionary[
        new_stereotype_uuid
    ] = stereotype_row_dictionary



from bclearer_core.common_knowledge.bclearer_additional_column_types import (
    BclearerAdditionalColumnTypes,
)
from bclearer_core.common_knowledge.bclearer_constants import (
    NAME_EXEMPLAR_ATTRIBUTE_NAME,
)
from bclearer_core.substages.operations.common.attribute_adder import (
    add_new_attribute_to_dictionary,
)
from bclearer_core.substages.operations.common.class_adder import (
    add_new_class_to_dictionary,
)
from bclearer_core.substages.operations.common.connector_adder import (
    add_new_connector_to_dictionary,
)
from bclearer_core.substages.operations.common.stereotype_adder import (
    add_new_stereotype_usage_to_dictionary,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.collection_types.nf_ea_com_collection_types import (
    NfEaComCollectionTypes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.column_types.nf_ea_com_column_types import (
    NfEaComColumnTypes,
)
from bclearer_interop_services.ea_interop_service.nf_ea_common.common_knowledge.ea_connector_types import (
    EaConnectorTypes,
)
from bclearer_interop_services.tuple_service.tuple_attribute_value_getter import (
    get_tuple_attribute_value_if_required,
)


def create_name_instance(
    new_ea_objects_dictionary: dict,
    package_nf_uuid: str,
    ea_attribute_tuple: tuple,
    name_instance_type_nf_uuid: str,
    name_types_instances_stereotype_nf_uuid: str,
    digitalisation_level_stereotype_nf_uuid: str,
):
    attributed_object_nf_uuid = get_tuple_attribute_value_if_required(
        owning_tuple=ea_attribute_tuple,
        attribute_name=NfEaComColumnTypes.ELEMENT_COMPONENTS_CONTAINING_EA_CLASSIFIER.column_name,
    )

    attributed_object_name = get_tuple_attribute_value_if_required(
        owning_tuple=ea_attribute_tuple,
        attribute_name=BclearerAdditionalColumnTypes.OWNING_OBJECT_NAMES.column_name,
    )

    name_instance_name = (
        attributed_object_name
        + " Instance"
    )

    name_instance_nf_uuid = add_new_class_to_dictionary(
        new_classifier_dictionary=new_ea_objects_dictionary[
            NfEaComCollectionTypes.EA_CLASSIFIERS
        ],
        package_nf_uuid=package_nf_uuid,
        class_name=name_instance_name,
    )

    add_new_stereotype_usage_to_dictionary(
        new_stereotype_usage_dictionary=new_ea_objects_dictionary[
            NfEaComCollectionTypes.STEREOTYPE_USAGE
        ],
        client_nf_uuid=name_instance_nf_uuid,
        client_collection_type=NfEaComCollectionTypes.EA_CLASSIFIERS,
        stereotype_nf_uuid=digitalisation_level_stereotype_nf_uuid,
    )

    add_new_connector_to_dictionary(
        new_connector_dictionary=new_ea_objects_dictionary[
            NfEaComCollectionTypes.EA_CONNECTORS
        ],
        place_1_nf_uuid=name_instance_nf_uuid,
        place_2_nf_uuid=name_instance_type_nf_uuid,
        connector_type=EaConnectorTypes.DEPENDENCY,
    )

    association_nf_uuid = add_new_connector_to_dictionary(
        new_connector_dictionary=new_ea_objects_dictionary[
            NfEaComCollectionTypes.EA_CONNECTORS
        ],
        place_1_nf_uuid=name_instance_nf_uuid,
        place_2_nf_uuid=attributed_object_nf_uuid,
        connector_type=EaConnectorTypes.ASSOCIATION,
    )

    add_new_stereotype_usage_to_dictionary(
        new_stereotype_usage_dictionary=new_ea_objects_dictionary[
            NfEaComCollectionTypes.STEREOTYPE_USAGE
        ],
        client_nf_uuid=association_nf_uuid,
        client_collection_type=NfEaComCollectionTypes.EA_CONNECTORS,
        stereotype_nf_uuid=name_types_instances_stereotype_nf_uuid,
    )

    attribute_value = get_tuple_attribute_value_if_required(
        owning_tuple=ea_attribute_tuple,
        attribute_name=NfEaComColumnTypes.ELEMENT_COMPONENTS_DEFAULT.column_name,
    )

    attribute_type_nf_uuid = get_tuple_attribute_value_if_required(
        owning_tuple=ea_attribute_tuple,
        attribute_name=NfEaComColumnTypes.ELEMENT_COMPONENTS_CLASSIFYING_EA_CLASSIFIER.column_name,
    )

    add_new_attribute_to_dictionary(
        new_ea_attributes_dictionary=new_ea_objects_dictionary[
            NfEaComCollectionTypes.EA_ATTRIBUTES
        ],
        name_object_nf_uuid=name_instance_nf_uuid,
        attribute_value=attribute_value,
        attribute_name=NAME_EXEMPLAR_ATTRIBUTE_NAME,
        attribute_type_nf_uuid=attribute_type_nf_uuid,
    )



from bclearer_core.common_knowledge.bclearer_additional_column_types import (
    BclearerAdditionalColumnTypes,
)
from bclearer_core.common_knowledge.bclearer_matched_ea_objects import (
    BclearerMatchedEaObjects,
)
from bclearer_core.common_knowledge.digitialisation_level_stereotype_matched_ea_objects import (
    DigitalisationLevelStereotypeMatchedEaObjects,
)
from bclearer_core.common_knowledge.matched_objects import (
    MatchedEaObjects,
)
from bclearer_core.constants.standard_constants import (
    DEFAULT_FOREIGN_TABLE_SUFFIX,
    DEFAULT_MASTER_TABLE_SUFFIX,
)
from bclearer_core.nf.types.nf_column_types import (
    NfColumnTypes,
)
from bclearer_core.substages.operations.b_evolve.convention_shift_operations.convention_shifters.separate_names_and_instances.ea_attribute_remover import (
    remove_ea_attributes,
)
from bclearer_core.substages.operations.b_evolve.convention_shift_operations.convention_shifters.separate_names_and_instances.name_instance_creator import (
    create_name_instance,
)
from bclearer_core.substages.operations.b_evolve.convention_shift_operations.convention_shifters.separate_names_and_instances.name_instance_type_creator import (
    create_name_instance_type,
)
from bclearer_core.substages.operations.common.instances_nf_uuids_getter import (
    get_instances_nf_uuids_of_matched_type,
)
from bclearer_core.substages.operations.common.new_ea_objects_dictionary_creator import (
    create_new_ea_objects_dictionary,
)
from bclearer_core.substages.operations.common.nf_ea_com_universe_updater import (
    update_nf_ea_com_universe_with_dictionary,
)
from bclearer_core.substages.operations.common.nf_uuid_from_ea_guid_from_collection_getter import (
    get_nf_uuid_from_ea_guid_from_collection,
)
from bclearer_interop_services.dataframe_service.dataframe_helpers.dataframe_mergers import (
    left_merge_dataframes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.collection_types.nf_ea_com_collection_types import (
    NfEaComCollectionTypes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.column_types.nf_ea_com_column_types import (
    NfEaComColumnTypes,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.nf_ea_com_universes import (
    NfEaComUniverses,
)
from bclearer_interop_services.ea_interop_service.nf_ea_common.common_knowledge.ea_connector_types import (
    EaConnectorTypes,
)
from bclearer_orchestration_services.reporting_service.wrappers.run_and_log_function_wrapper import (
    run_and_log_function,
)
from pandas import DataFrame


@run_and_log_function
def separate_names_and_instances(
    nf_ea_com_universe: NfEaComUniverses,
    matched_naming_space_type: MatchedEaObjects,
    name_instance_attribute_name: str,
    package_nf_uuid: str,
):
    ea_connectors = (
        nf_ea_com_universe.nf_ea_com_registry.get_ea_connectors()
    )

    ea_dependencies = ea_connectors[
        ea_connectors[
            NfEaComColumnTypes.CONNECTORS_ELEMENT_TYPE_NAME.column_name
        ]
        == EaConnectorTypes.DEPENDENCY.type_name
    ]

    new_ea_objects_dictionary = (
        create_new_ea_objects_dictionary()
    )

    naming_space_instances = get_instances_nf_uuids_of_matched_type(
        nf_ea_com_universe=nf_ea_com_universe,
        matched_type=matched_naming_space_type,
    )

    __remove_attribute_from_type(
        nf_ea_com_universe=nf_ea_com_universe,
        matched_naming_space_type=matched_naming_space_type,
        name_instance_attribute_name=name_instance_attribute_name,
    )

    digitalisation_level_stereotype_nf_uuid = get_nf_uuid_from_ea_guid_from_collection(
        nf_ea_com_universe=nf_ea_com_universe,
        collection_type=NfEaComCollectionTypes.EA_STEREOTYPES,
        ea_guid=DigitalisationLevelStereotypeMatchedEaObjects.DIGITALISATION_LEVEL_1_CLASS_STEREOTYPE.ea_guid,
    )

    for (
        naming_space_instance_nf_uuid
    ) in naming_space_instances:
        __separate_name_instances(
            nf_ea_com_universe=nf_ea_com_universe,
            ea_dependencies=ea_dependencies,
            naming_space_instance_nf_uuid=naming_space_instance_nf_uuid,
            name_instance_attribute_name=name_instance_attribute_name,
            new_ea_objects_dictionary=new_ea_objects_dictionary,
            package_nf_uuid=package_nf_uuid,
            digitalisation_level_stereotype_nf_uuid=digitalisation_level_stereotype_nf_uuid,
        )

    update_nf_ea_com_universe_with_dictionary(
        nf_ea_com_universe=nf_ea_com_universe,
        new_ea_objects_dictionary=new_ea_objects_dictionary,
    )


@run_and_log_function
def __separate_name_instances(
    nf_ea_com_universe: NfEaComUniverses,
    ea_dependencies: DataFrame,
    naming_space_instance_nf_uuid: str,
    name_instance_attribute_name: str,
    new_ea_objects_dictionary: dict,
    package_nf_uuid: str,
    digitalisation_level_stereotype_nf_uuid: str,
):
    name_instance_type_nf_uuid = create_name_instance_type(
        nf_ea_com_universe=nf_ea_com_universe,
        new_ea_objects_dictionary=new_ea_objects_dictionary,
        package_nf_uuid=package_nf_uuid,
        name_object_name=name_instance_attribute_name,
        naming_space_nf_uuid=naming_space_instance_nf_uuid,
        digitalisation_level_stereotype_nf_uuid=digitalisation_level_stereotype_nf_uuid,
    )

    name_types_instances_stereotype_nf_uuid = get_nf_uuid_from_ea_guid_from_collection(
        nf_ea_com_universe=nf_ea_com_universe,
        collection_type=NfEaComCollectionTypes.EA_STEREOTYPES,
        ea_guid=BclearerMatchedEaObjects.NAME_TYPES_INSTANCES_STEREOTYPE.ea_guid,
    )

    name_instance_attributes = __get_name_instance_attributes(
        nf_ea_com_universe=nf_ea_com_universe,
        ea_dependencies=ea_dependencies,
        naming_space_instance_nf_uuid=naming_space_instance_nf_uuid,
        name_instance_attribute_name=name_instance_attribute_name,
    )

    for (
        ea_attribute_tuple
    ) in (
        name_instance_attributes.itertuples()
    ):
        create_name_instance(
            new_ea_objects_dictionary=new_ea_objects_dictionary,
            package_nf_uuid=package_nf_uuid,
            ea_attribute_tuple=ea_attribute_tuple,
            name_instance_type_nf_uuid=name_instance_type_nf_uuid,
            name_types_instances_stereotype_nf_uuid=name_types_instances_stereotype_nf_uuid,
            digitalisation_level_stereotype_nf_uuid=digitalisation_level_stereotype_nf_uuid,
        )

    remove_ea_attributes(
        nf_ea_com_universe=nf_ea_com_universe,
        ea_attributes=name_instance_attributes,
    )


@run_and_log_function
def __get_name_instance_attributes(
    nf_ea_com_universe: NfEaComUniverses,
    ea_dependencies: DataFrame,
    naming_space_instance_nf_uuid: str,
    name_instance_attribute_name: str,
) -> DataFrame:
    filtered_dependencies = ea_dependencies[
        ea_dependencies[
            NfEaComColumnTypes.ELEMENTS_CLIENT_PLACE2_END_CONNECTORS.column_name
        ]
        == naming_space_instance_nf_uuid
    ]

    name_instances = set(
        filtered_dependencies[
            NfEaComColumnTypes.ELEMENTS_SUPPLIER_PLACE1_END_CONNECTORS.column_name
        ],
    )

    ea_attributes = (
        nf_ea_com_universe.nf_ea_com_registry.get_ea_attributes()
    )

    name_instance_attributes = ea_attributes.loc[
        ea_attributes[
            NfEaComColumnTypes.ELEMENT_COMPONENTS_CONTAINING_EA_CLASSIFIER.column_name
        ].isin(name_instances)
    ]

    filtered_name_instance_attributes = name_instance_attributes[
        name_instance_attributes[
            NfEaComColumnTypes.EXPLICIT_OBJECTS_EA_OBJECT_NAME.column_name
        ]
        == name_instance_attribute_name
    ]

    ea_classifiers = (
        nf_ea_com_universe.nf_ea_com_registry.get_ea_classifiers()
    )

    extended_filtered_name_instance_attributes = left_merge_dataframes(
        master_dataframe=filtered_name_instance_attributes,
        master_dataframe_key_columns=[
            NfEaComColumnTypes.ELEMENT_COMPONENTS_CONTAINING_EA_CLASSIFIER.column_name,
        ],
        merge_suffixes=[
            DEFAULT_FOREIGN_TABLE_SUFFIX,
            DEFAULT_MASTER_TABLE_SUFFIX,
        ],
        foreign_key_dataframe=ea_classifiers,
        foreign_key_dataframe_fk_columns=[
            NfColumnTypes.NF_UUIDS.column_name,
        ],
        foreign_key_dataframe_other_column_rename_dictionary={
            NfEaComColumnTypes.EXPLICIT_OBJECTS_EA_OBJECT_NAME.column_name: BclearerAdditionalColumnTypes.OWNING_OBJECT_NAMES.column_name,
        },
    )

    return extended_filtered_name_instance_attributes


def __remove_attribute_from_type(
    nf_ea_com_universe: NfEaComUniverses,
    matched_naming_space_type: MatchedEaObjects,
    name_instance_attribute_name: str,
):
    naming_space_type_nf_uuid = get_nf_uuid_from_ea_guid_from_collection(
        nf_ea_com_universe=nf_ea_com_universe,
        collection_type=NfEaComCollectionTypes.EA_STEREOTYPES,
        ea_guid=matched_naming_space_type.ea_guid,
    )

    ea_attributes = (
        nf_ea_com_universe.nf_ea_com_registry.get_ea_attributes()
    )

    naming_space_type_attributes = ea_attributes.loc[
        ea_attributes[
            NfEaComColumnTypes.ELEMENT_COMPONENTS_CONTAINING_EA_CLASSIFIER.column_name
        ]
        == naming_space_type_nf_uuid
    ]

    filtered_naming_space_type_attributes = naming_space_type_attributes[
        naming_space_type_attributes[
            NfEaComColumnTypes.EXPLICIT_OBJECTS_EA_OBJECT_NAME.column_name
        ]
        == name_instance_attribute_name
    ]

    remove_ea_attributes(
        nf_ea_com_universe=nf_ea_com_universe,
        ea_attributes=filtered_naming_space_type_attributes,
    )



from enum import auto, unique

from bclearer_core.common_knowledge.matched_objects import (
    MatchedEaObjects,
)


@unique
class DigitalisationLevelStereotypeMatchedEaObjects(
    MatchedEaObjects,
):
    DIGITALISATION_LEVEL_1_CLASS_STEREOTYPE = (
        auto()
    )

    DIGITALISATION_LEVEL_2_CLASS_STEREOTYPE = (
        auto()
    )

    DIGITALISATION_LEVEL_3_CLASS_STEREOTYPE = (
        auto()
    )

    DIGITALISATION_LEVEL_4_CLASS_STEREOTYPE = (
        auto()
    )

    DIGITALISATION_LEVEL_5_CLASS_STEREOTYPE = (
        auto()
    )

    DIGITALISATION_LEVEL_6_CLASS_STEREOTYPE = (
        auto()
    )

    DIGITALISATION_LEVEL_7_CLASS_STEREOTYPE = (
        auto()
    )

    DIGITALISATION_LEVEL_8_CLASS_STEREOTYPE = (
        auto()
    )

    @classmethod
    def get_ea_guids(cls) -> set:
        ea_guids = set(
            ea_guid_mapping.values(),
        )

        return ea_guids

    def __object_name(self) -> str:
        object_name = (
            object_name_mapping[self]
        )

        return object_name

    def __ea_guid(self) -> str:
        ea_guid = ea_guid_mapping[self]

        return ea_guid

    def __style(self) -> str:
        style = (
            digitalisation_level_styles[
                self
            ]
        )

        return style

    object_name = property(
        fget=__object_name,
    )

    ea_guid = property(fget=__ea_guid)

    style = property(fget=__style)


object_name_mapping = {
    DigitalisationLevelStereotypeMatchedEaObjects.DIGITALISATION_LEVEL_1_CLASS_STEREOTYPE: "DL1",
    DigitalisationLevelStereotypeMatchedEaObjects.DIGITALISATION_LEVEL_2_CLASS_STEREOTYPE: "DL2",
    DigitalisationLevelStereotypeMatchedEaObjects.DIGITALISATION_LEVEL_3_CLASS_STEREOTYPE: "DL3",
    DigitalisationLevelStereotypeMatchedEaObjects.DIGITALISATION_LEVEL_4_CLASS_STEREOTYPE: "DL4",
    DigitalisationLevelStereotypeMatchedEaObjects.DIGITALISATION_LEVEL_5_CLASS_STEREOTYPE: "DL5",
    DigitalisationLevelStereotypeMatchedEaObjects.DIGITALISATION_LEVEL_6_CLASS_STEREOTYPE: "DL6",
    DigitalisationLevelStereotypeMatchedEaObjects.DIGITALISATION_LEVEL_7_CLASS_STEREOTYPE: "DL7",
    DigitalisationLevelStereotypeMatchedEaObjects.DIGITALISATION_LEVEL_8_CLASS_STEREOTYPE: "DL8",
}


ea_guid_mapping = {
    DigitalisationLevelStereotypeMatchedEaObjects.DIGITALISATION_LEVEL_1_CLASS_STEREOTYPE: "{FF73ABA9-DE5C-4eeb-918C-DECA01FF7280}",
    DigitalisationLevelStereotypeMatchedEaObjects.DIGITALISATION_LEVEL_2_CLASS_STEREOTYPE: "{0AA699E0-7794-4aea-8F27-EED1706BB74D}",
    DigitalisationLevelStereotypeMatchedEaObjects.DIGITALISATION_LEVEL_3_CLASS_STEREOTYPE: "{2A9FD9A6-9D8F-421c-BF26-A3FBA21268B4}",
    DigitalisationLevelStereotypeMatchedEaObjects.DIGITALISATION_LEVEL_4_CLASS_STEREOTYPE: "{24129ADB-A8B8-42e8-8EBB-2954690FB64D}",
    DigitalisationLevelStereotypeMatchedEaObjects.DIGITALISATION_LEVEL_5_CLASS_STEREOTYPE: "{EE83A38C-E442-498b-A594-26699B46DC0F}",
    DigitalisationLevelStereotypeMatchedEaObjects.DIGITALISATION_LEVEL_6_CLASS_STEREOTYPE: "{5D0561A8-ECB7-4016-AD3F-8F80135F5CB9}",
    DigitalisationLevelStereotypeMatchedEaObjects.DIGITALISATION_LEVEL_7_CLASS_STEREOTYPE: "{5BD3B142-98DF-4c94-B740-D000DCB4A009}",
    DigitalisationLevelStereotypeMatchedEaObjects.DIGITALISATION_LEVEL_8_CLASS_STEREOTYPE: "{8A1027F5-6CA2-49d8-A933-1DF26D07642C}",
}


digitalisation_level_styles = {
    DigitalisationLevelStereotypeMatchedEaObjects.DIGITALISATION_LEVEL_1_CLASS_STEREOTYPE: '<STYLE fill="1837750" text="-1" border="-1" groupname="Digitalisation Level Group" type="metafile"/>',
    DigitalisationLevelStereotypeMatchedEaObjects.DIGITALISATION_LEVEL_2_CLASS_STEREOTYPE: '<STYLE fill="3225056" text="-1" border="-1" groupname="Digitalisation Level Group" type="metafile"/>',
    DigitalisationLevelStereotypeMatchedEaObjects.DIGITALISATION_LEVEL_3_CLASS_STEREOTYPE: '<STYLE fill="5007615" text="-1" border="-1" groupname="Digitalisation Level Group" type="metafile"/>',
    DigitalisationLevelStereotypeMatchedEaObjects.DIGITALISATION_LEVEL_4_CLASS_STEREOTYPE: '<STYLE fill="170211" text="-1" border="-1" groupname="Digitalisation Level Group" type="metafile"/>',
    DigitalisationLevelStereotypeMatchedEaObjects.DIGITALISATION_LEVEL_5_CLASS_STEREOTYPE: '<STYLE fill="2604528" text="-1" border="-1" groupname="Digitalisation Level Group" type="metafile"/>',
    DigitalisationLevelStereotypeMatchedEaObjects.DIGITALISATION_LEVEL_6_CLASS_STEREOTYPE: '<STYLE fill="6740735" text="-1" border="-1" groupname="Digitalisation Level Group" type="metafile"/>',
    DigitalisationLevelStereotypeMatchedEaObjects.DIGITALISATION_LEVEL_7_CLASS_STEREOTYPE: '<STYLE fill="8310410" text="-1" border="-1" groupname="Digitalisation Level Group" type="metafile"/>',
    DigitalisationLevelStereotypeMatchedEaObjects.DIGITALISATION_LEVEL_8_CLASS_STEREOTYPE: '<STYLE fill="4428080" text="-1" border="-1" groupname="Digitalisation Level Group" type="metafile"/>',
}


digitalisation_level_order = {
    DigitalisationLevelStereotypeMatchedEaObjects.DIGITALISATION_LEVEL_1_CLASS_STEREOTYPE: 1,
    DigitalisationLevelStereotypeMatchedEaObjects.DIGITALISATION_LEVEL_2_CLASS_STEREOTYPE: 2,
    DigitalisationLevelStereotypeMatchedEaObjects.DIGITALISATION_LEVEL_3_CLASS_STEREOTYPE: 3,
    DigitalisationLevelStereotypeMatchedEaObjects.DIGITALISATION_LEVEL_4_CLASS_STEREOTYPE: 4,
    DigitalisationLevelStereotypeMatchedEaObjects.DIGITALISATION_LEVEL_5_CLASS_STEREOTYPE: 5,
    DigitalisationLevelStereotypeMatchedEaObjects.DIGITALISATION_LEVEL_6_CLASS_STEREOTYPE: 6,
    DigitalisationLevelStereotypeMatchedEaObjects.DIGITALISATION_LEVEL_7_CLASS_STEREOTYPE: 7,
    DigitalisationLevelStereotypeMatchedEaObjects.DIGITALISATION_LEVEL_8_CLASS_STEREOTYPE: 8,
}



class LoadEaModelConfigurations:
    def __init__(
        self,
        resource_namespace: str,
        resource_name: str,
        short_name: str,
    ):
        self.resource_namespace = (
            resource_namespace
        )

        self.resource_name = (
            resource_name
        )

        self.short_name = short_name

    def __enter__(self):
        return self

    def __exit__(
        self,
        exception_type,
        exception_value,
        traceback,
    ):
        pass



from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.collection_types.nf_ea_com_collection_types import (
    NfEaComCollectionTypes,
)
from bclearer_orchestration_services.reporting_service.reporters.log_with_datetime import (
    log_message,
)


def report_universes_with_different_collection_types(
    collection_types_in_primary_universe: set,
    collection_types_in_aligned_universe: set,
):
    primary_to_aligned_delta = collection_types_in_primary_universe.difference(
        collection_types_in_aligned_universe,
    )

    aligned_to_primary_delta = collection_types_in_aligned_universe.difference(
        collection_types_in_primary_universe,
    )

    log_message(
        message="Universes cannot be fully aligned because they keep different collections.",
    )

    if (
        len(primary_to_aligned_delta)
        > 0
    ):
        log_message(
            message="Primary minus aligned "
            + repr(
                primary_to_aligned_delta,
            ),
        )

    if (
        len(aligned_to_primary_delta)
        > 0
    ):
        log_message(
            message="Aligned minus primary "
            + repr(
                aligned_to_primary_delta,
            ),
        )


def report_collection_with_different_columns(
    primary_universe_collection_columns: set,
    aligned_universe_collection_columns: set,
    collection_type: NfEaComCollectionTypes,
):
    primary_to_aligned_delta = primary_universe_collection_columns.difference(
        aligned_universe_collection_columns,
    )

    aligned_to_primary_delta = aligned_universe_collection_columns.difference(
        primary_universe_collection_columns,
    )

    log_message(
        message=collection_type.collection_name
        + " collection type has different columns in the universes being merged.",
    )

    if (
        len(primary_to_aligned_delta)
        > 0
    ):
        log_message(
            message="Primary minus aligned "
            + repr(
                primary_to_aligned_delta,
            ),
        )

    if (
        len(aligned_to_primary_delta)
        > 0
    ):
        log_message(
            message="Aligned minus primary "
            + repr(
                aligned_to_primary_delta,
            ),
        )



import importlib
from pathlib import Path

from bclearer_core.configurations.load_hdf5_model_configurations import (
    LoadHdf5ModelConfigurations,
)
from bclearer_core.substages.operations.a_load.content_operations.digitalisation_levels.default_digitalisation_level_stereotype_adder import (
    add_default_digitalisation_level_stereotype,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.nf_ea_com_universes import (
    NfEaComUniverses,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.model_loader.hdf5_loader.hdf5_to_nf_ea_com_universe_loader import (
    load_hdf5_to_nf_ea_com_universe,
)
from bclearer_interop_services.ea_interop_service.session.orchestrators.ea_tools_session_managers import (
    EaToolsSessionManagers,
)
from bclearer_interop_services.file_system_service.objects.files import (
    Files,
)
from bclearer_orchestration_services.reporting_service.reporters.log_with_datetime import (
    log_message,
)
from bclearer_orchestration_services.reporting_service.wrappers.run_and_log_function_wrapper import (
    run_and_log_function,
)


@run_and_log_function
def load_hdf5_model_to_content_universe(
    ea_tools_session_manager: EaToolsSessionManagers,
    load_hdf5_model_configuration: LoadHdf5ModelConfigurations,
) -> NfEaComUniverses:
    log_message(
        message="CONTENT OPERATION: Load universe from HDF5 - "
        + load_hdf5_model_configuration.universe_short_name
        + " - started",
    )

    hdf5_file = __get_hdf5_file(
        resource_namespace=load_hdf5_model_configuration.resource_namespace,
        resource_file_name=load_hdf5_model_configuration.resource_file_name,
    )

    output_universe = load_hdf5_to_nf_ea_com_universe(
        ea_tools_session_manager=ea_tools_session_manager,
        hdf5_file=hdf5_file,
        short_name=load_hdf5_model_configuration.universe_short_name,
    )

    if (
        load_hdf5_model_configuration.default_digitalisation_level_stereotype
    ):
        add_default_digitalisation_level_stereotype(
            nf_ea_com_universe=output_universe,
            default_digitalisation_level_stereotype=load_hdf5_model_configuration.default_digitalisation_level_stereotype,
        )

    log_message(
        message="CONTENT OPERATION: Load universe from HDF5 - "
        + load_hdf5_model_configuration.universe_short_name
        + " - finished",
    )

    return output_universe


def __get_empty_nf_ea_com_universe(
    ea_tools_session_manager: EaToolsSessionManagers,
    short_name: str,
) -> NfEaComUniverses:
    ea_repository = ea_tools_session_manager.create_empty_ea_repository_with_short_name(
        short_name=short_name,
    )

    nf_ea_com_universe_manager = (
        ea_tools_session_manager.nf_ea_com_endpoint_manager.nf_ea_com_universe_manager
    )

    nf_ea_com_universe = nf_ea_com_universe_manager.nf_ea_com_universe_dictionary[
        ea_repository
    ]

    return nf_ea_com_universe


def __get_hdf5_file(
    resource_namespace: str,
    resource_file_name: str,
) -> Files:
    resource_namespace_module = (
        importlib.import_module(
            name=resource_namespace,
        )
    )

    resource_namespace_folder_name = resource_namespace_module.__path__[
        0
    ]

    resource_folder_path = Path(
        resource_namespace_folder_name,
    )

    hdf5_file_name = (
        resource_folder_path.joinpath(
            resource_file_name,
        )
    )

    hdf5_file = Files(
        absolute_path_string=str(
            hdf5_file_name,
        ),
    )

    return hdf5_file



from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.collection_types.nf_ea_com_collection_types import (
    NfEaComCollectionTypes,
)

LIST_OF_COLLECTION_TYPES_OF_OBJECTS_WITHOUT_EA_GUIDS = [
    NfEaComCollectionTypes.EA_CARDINALITIES,
    NfEaComCollectionTypes.EA_CONNECTOR_TYPES,
    NfEaComCollectionTypes.EA_DIAGRAM_TYPES,
    NfEaComCollectionTypes.EA_ELEMENT_TYPES,
]



def convert_list_of_sets_to_list_of_lists(
    list_of_sets: list,
) -> list:
    list_of_lists = []

    for set_1 in list_of_sets:
        __convert_set_to_list_and_append(
            list_of_lists,
            set_1,
        )

    return list_of_lists


def __convert_set_to_list_and_append(
    list_of_lists: list,
    set_1: set,
):
    set_as_list = []

    for element in set_1:
        set_as_list.append(element)

    list_of_lists.append(set_as_list)



from bclearer_core.nf.types.nf_column_types import (
    NfColumnTypes,
)
from bclearer_core.substages.operations.common.ea_guid_from_nf_uuid_creator import (
    create_ea_guid_from_nf_uuid,
)
from bclearer_interop_services.ea_interop_service.general.nf_ea.com.common_knowledge.column_types.nf_ea_com_column_types import (
    NfEaComColumnTypes,
)
from bclearer_interop_services.ea_interop_service.nf_ea_common.common_knowledge.ea_connector_types import (
    EaConnectorTypes,
)
from bclearer_orchestration_services.identification_services.uuid_service.uuid_helpers.uuid_factory import (
    create_new_uuid,
)


def create_connector(
    new_ea_connectors_dictionary: dict,
    place_1_nf_uuid: str,
    place_2_nf_uuid: str,
    connector_name: str,
    connector_type: EaConnectorTypes,
) -> str:
    new_connector_nf_uuid = (
        create_new_uuid()
    )

    new_connector_ea_guid = create_ea_guid_from_nf_uuid(
        nf_uuid=new_connector_nf_uuid,
    )

    nf_ea_com_universe_new_connector_row_dictionary = {
        NfColumnTypes.NF_UUIDS.column_name: new_connector_nf_uuid,
        NfEaComColumnTypes.ELEMENTS_SUPPLIER_PLACE1_END_CONNECTORS.column_name: place_1_nf_uuid,
        NfEaComColumnTypes.ELEMENTS_CLIENT_PLACE2_END_CONNECTORS.column_name: place_2_nf_uuid,
        NfEaComColumnTypes.CONNECTORS_DIRECTION_TYPE_NAME.column_name: "Source -> Destination",
        NfEaComColumnTypes.CONNECTORS_ELEMENT_TYPE_NAME.column_name: connector_type.type_name,
        NfEaComColumnTypes.STEREOTYPEABLE_OBJECTS_EA_OBJECT_STEREOTYPES.column_name: [],
        NfEaComColumnTypes.EXPLICIT_OBJECTS_EA_OBJECT_NAME.column_name: connector_name,
        NfEaComColumnTypes.EXPLICIT_OBJECTS_EA_GUID.column_name: new_connector_ea_guid,
    }

    new_ea_connectors_dictionary[
        len(
            new_ea_connectors_dictionary,
        )
        + 1
    ] = nf_ea_com_universe_new_connector_row_dictionary

    return new_connector_nf_uuid
